class QuantityValidationError(ValueError):
    def __init__(
        self, message: str, caption: str, value: float, operator: str, limit_value: float
    ) -> None:
        """
        Value error subclass with helpers attributes.

        :param message:
            the full formatted message
        :param caption:
            caption of the category info
        :param value:
            value validated
        :param operator:
            operator (literal or not)
        :param limit_value:
            limit value
        """
        super().__init__(message)
        self.message = message
        self.caption = caption
        self.value = value
        self.operator = operator
        self.limit_value = limit_value



import copy

import pytest

from barril.basic.fraction import Fraction
from barril.basic.fraction import FractionValue


def testBasicUsage() -> None:
    f = FractionValue(3, Fraction(5, 3))
    assert f.number == 3
    assert f.fraction == Fraction(5, 3)

    f.number = 5.5
    f.fraction = Fraction(6, 5)
    assert f.number == 5.5
    assert f.fraction == Fraction(6, 5)

    with pytest.raises(TypeError):
        f.SetNumber("hello")  # type:ignore[arg-type]
    with pytest.raises(TypeError):
        f.SetFraction("hello")  # type:ignore[arg-type]
    with pytest.raises(ValueError):
        f.SetFraction((1, 2, 3))  # type:ignore[arg-type]

    assert FractionValue(3).GetFraction() == Fraction(0, 1)


def testDefault() -> None:
    f = FractionValue()
    assert f.number == 0.0
    assert f.fraction == Fraction(0, 1)


def testPartsArentNone() -> None:
    """
    FractionValue can't be initialized nor modified to have None as number or fraction part.
    """
    with pytest.raises(TypeError):
        FractionValue(1, None)  # type:ignore[arg-type]
    with pytest.raises(TypeError):
        FractionValue(None, (0 / 1))  # type:ignore[arg-type]
    with pytest.raises(TypeError):
        FractionValue(None, None)  # type:ignore[arg-type]

    f = FractionValue(1, Fraction(0, 1))
    with pytest.raises(TypeError):
        f.SetNumber(None)  # type:ignore[arg-type]
    with pytest.raises(TypeError):
        f.SetFraction(None)  # type:ignore[arg-type]


def testMatchFractionPart() -> None:
    # Text Ok, should not raise error
    FractionValue.MatchFractionPart("3/4")
    with pytest.raises(ValueError):
        FractionValue.MatchFractionPart("2 3/4")


def testStr() -> None:
    f = FractionValue(3, Fraction(5, 3))
    assert str(f) == "3 5/3"
    assert repr(f) == "FractionValue(3, 5/3)"

    f = FractionValue(3)
    assert str(f) == "3"
    assert repr(f) == "FractionValue(3, 0/1)"


def testEquality() -> None:
    assert FractionValue(3, Fraction(5, 3)) == FractionValue(3, Fraction(5, 3))
    assert not FractionValue(3, Fraction(5, 3)) != FractionValue(3, Fraction(5, 3))

    assert FractionValue(3) == FractionValue(3)
    assert not FractionValue(3) != FractionValue(3)

    assert FractionValue(10, Fraction(5, 3)) != FractionValue(3, Fraction(5, 3))
    assert not FractionValue(10, Fraction(5, 3)) == FractionValue(3, Fraction(5, 3))

    assert FractionValue(10, (5, 3)) == FractionValue(10, Fraction(5, 3))


def testFloat() -> None:
    assert float(FractionValue(3, Fraction(5, 3))) == 3 + 5 / 3.0
    assert float(FractionValue(3)) == 3.0


def testCopy() -> None:
    f = FractionValue(3, (5, 3))
    cf = copy.copy(f)
    assert f == cf
    assert f is not cf

    cf.fraction.numerator = 10
    cf.fraction.denominator = 4
    assert f == FractionValue(3, (5, 3))


def testComparison() -> None:
    assert FractionValue(3) < FractionValue(3, (3, 4))
    assert FractionValue(3) <= FractionValue(3, (3, 4))
    assert FractionValue(3, (3, 4)) > FractionValue(3)
    assert FractionValue(3, (3, 4)) >= FractionValue(3)


def testCreateFromString() -> None:
    """
    Allow the user enter only the fraction value
    """

    def AssertCreateFromString(text, whole, fraction=None):
        assert (
            FractionValue.CreateFromString(text) == FractionValue(whole, fraction)
            if fraction is not None
            else FractionValue(whole)
        )

    AssertCreateFromString("1 1/2", 1, (1, 2))
    AssertCreateFromString("1", 1)
    AssertCreateFromString("1.2 1.1/3", 1.2, (1.1, 3))
    AssertCreateFromString("1.2", 1.2)
    AssertCreateFromString("1/3", 0, (1, 3))
    AssertCreateFromString("3.3/4", 0, (3.3, 4))

    AssertCreateFromString("33/4", 0, (33, 4))
    AssertCreateFromString("  35/4", 0, (35, 4))
    AssertCreateFromString("3 36/4", 3, (36, 4))


def testCreateFromFloat() -> None:
    """
    Allow the user enter only the fraction value
    """
    assert FractionValue.CreateFromFloat(1.375) == FractionValue(1, (3, 8))
    assert FractionValue.CreateFromFloat(2.5) == FractionValue(2, (1, 2))



import pytest
from pytest import approx

from barril import units
from barril.basic.format_float import FormatFloat
from barril.basic.fraction import FractionValue
from barril.units import Array
from barril.units import FixedArray
from barril.units import InvalidQuantityTypeError
from barril.units import InvalidUnitError
from barril.units import ObtainQuantity
from barril.units import Quantity
from barril.units import UnitsError
from barril.units import UNKNOWN_QUANTITY_TYPE
from barril.units import UNKNOWN_UNIT
from barril.units.exceptions import QuantityValidationError
from barril.units.unit_database import UnitDatabase


def testNotDefaultUnitDatabase() -> None:
    unit_database = units.UnitDatabase()
    with pytest.raises(AssertionError):
        unit_database.CheckDefaultUnitDatabase()


def testDefaultUnitDatabase(unit_database) -> None:
    unit_database.CheckDefaultUnitDatabase()


def testConv(unit_database_empty) -> None:
    unit_database = unit_database_empty

    # add some units for testing
    unit_database.AddUnitBase("temperature", "Celcius", "ºC")
    unit_database.AddUnit("temperature", "Fahrenheit", "F", "%f * 1.8 + 32.0", " (%f - 32.0) / 1.8")
    unit_database.AddUnit("temperature", "Kelvin", "K", "%f + 273.15", "%f - 273.15")
    unit_database.AddUnit(
        "temperature", "Rakini", "R", "%f * 1.8 + 32 + 459.67", "(%f - 32 - 459.67) / 1.8"
    )

    assert 50 == unit_database.Convert("temperature", "ºC", "F", 10)
    assert 10 == unit_database.Convert("temperature", "F", "ºC", 50)
    assert approx(abs(-245.372222 - unit_database.Convert("temperature", "R", "ºC", 50)), 5) == 0
    assert approx(abs(581.67 - unit_database.Convert("temperature", "ºC", "R", 50)), 7) == 0

    assert [50, 50, 50] == unit_database.Convert("temperature", "ºC", "F", [10, 10, 10])
    assert [10, 10, 10] == unit_database.Convert("temperature", "F", "ºC", [50, 50, 50])
    assert (10, 10, 10) == unit_database.Convert("temperature", "F", "ºC", (50, 50, 50))


def testPrecision() -> None:
    unit_database = units.UnitDatabase()
    unit_database.AddUnitBase("Compressibility", "1/Bars", "1/Bars")
    unit_database.AddUnit(
        "Compressibility", "1/Psi", "1/Psi", frombase="%f / 14.50377", tobase="%f * 14.50377"
    )
    unit_database.AddUnit(
        "Compressibility", "1/Atm", "1/Atm", frombase="%f / 0.986923", tobase="%f * 0.986923"
    )

    assert (
        approx(
            abs(
                14.50377 * 10e-6
                - unit_database.Convert("Compressibility", "1/Psi", "1/Bars", 1 * 10e-6)
            ),
            5,
        )
        == 0
    )
    assert (
        approx(
            abs(
                0.000063816588
                - unit_database.Convert("Compressibility", "1/Psi", "1/Bars", 4.4e-006)
            ),
            5,
        )
        == 0
    )


def testQuantityTypes(unit_database_custom_conversion) -> None:
    unit_database = unit_database_custom_conversion
    real = unit_database.GetQuantityTypes()
    real.sort()
    assert real == [UNKNOWN_QUANTITY_TYPE, "length", "temperature"]


def testUnitQuantityType(unit_database_custom_conversion) -> None:
    unit_database = unit_database_custom_conversion
    assert "length" == unit_database.GetQuantityType("m")


def testDistanceUnits(unit_database_custom_conversion) -> None:
    unit_database = unit_database_custom_conversion
    available_units = ["m", "mm", "cm", "km", "mi", "in", "µm"]
    real = unit_database.GetUnits("length")
    assert available_units == real


def testFindCase(unit_database_custom_conversion) -> None:
    unit_database = units.UnitDatabase()
    # add some units for testing
    unit_database.AddUnitBase("temperature", "mang1", "mA")
    unit_database.AddUnit("temperature", "mang2", "Ma", frombase="%f", tobase="%f")
    unit_database.AddUnit("temperature", "Celsius", "C", frombase="%f", tobase="%f")
    unit_database.AddUnit("temperature", "mang3", "MA", frombase="%f", tobase="%f")

    unit_database.AddCategory("my", "temperature")
    with pytest.raises(AssertionError):
        unit_database.FindUnitCase("my", "ma")
    assert "C" == unit_database.FindUnitCase("my", "c")
    assert "C" == unit_database.FindUnitCase("my", "C")


def testUnits(unit_database_custom_conversion) -> None:
    unit_database = unit_database_custom_conversion
    real = unit_database.GetUnits()
    real.sort()
    available_units = [UNKNOWN_UNIT, "m", "mm", "cm", "km", "mi", "in", "µm", "ºC", "F", "K"]
    available_units.sort()
    assert available_units == real


def testBaseUnit(unit_database_custom_conversion) -> None:
    unit_database = unit_database_custom_conversion
    assert "m" == unit_database.GetBaseUnit("length")


def testUnitNames(unit_database_custom_conversion) -> None:
    unit_database = unit_database_custom_conversion
    names = {"meters", "milimeters", "centimeters", "kilometers", "miles", "inches", "micrometers"}
    onames = set(unit_database.GetUnitNames("length"))
    assert names == onames


def testClear(unit_database_custom_conversion) -> None:
    unit_database = unit_database_custom_conversion
    unit_database.Clear()
    with pytest.raises(units.InvalidQuantityTypeError):
        unit_database.CheckQuantityType("length")


def testCategory(unit_database_custom_conversion) -> None:
    unit_database = unit_database_custom_conversion
    unit_database.AddCategory(
        "my length",
        "length",
        ["mm", "m"],
        default_unit="m",
        default_value=15.5,
        min_value=-6e-10,
        max_value=2e5,
        is_min_exclusive=True,
        is_max_exclusive=False,
        caption="My Length",
    )

    assert unit_database.GetDefaultValue("my length") == 15.5
    assert unit_database.GetDefaultUnit("my length") == "m"
    assert sorted(list(unit_database.IterCategories())) == ["length", "my length"]

    quantity = ObtainQuantity("m", "my length")
    formatted_value = FormatFloat("%g", -6e-010)
    with pytest.raises(
        QuantityValidationError,
        match="Invalid value for My Length: %s. Must be > %s." % (formatted_value, formatted_value),
    ) as exc_info:
        quantity.CheckValue(-6e-10)
    e = exc_info.value
    assert e.message == "Invalid value for My Length: {}. Must be > {}.".format(
        formatted_value, formatted_value
    )
    assert e.caption == "My Length"
    assert e.value == float(formatted_value)
    assert e.operator == ">"
    assert e.value == float(formatted_value)

    quantity.CheckValue(0)  # without specifying unit
    mm_quantity = ObtainQuantity("mm", "my length")
    mm_quantity.CheckValue(2e5)

    with pytest.raises(
        QuantityValidationError, match="Invalid value for My Length: 200000. Must be <= 200000.0."
    ):
        mm_quantity.CheckValue(2e8 + 1)

    # Check the unit info using a category instead a quantity_type
    with pytest.raises(InvalidQuantityTypeError):
        unit_database.GetInfo("unknown length", "mm")

    unit_database.GetInfo("my length", "mm")
    unit_database.GetInfo(UNKNOWN_QUANTITY_TYPE, "I know it is m3", fix_unknown=True)


def testUniqueness(unit_database_custom_conversion) -> None:
    # trying to re-register milimeters (mm)
    unit_database = unit_database_custom_conversion
    with pytest.raises(RuntimeError):
        unit_database.AddUnit("length", "milimeters", "mm", "%f * 1000.0", "%f / 1000.0")


def testDefaultUnitWhenNoneIsPassed(unit_database_custom_conversion) -> None:
    unit_database = unit_database_custom_conversion
    unit_database.AddCategory("my category", "length", ["mm"], default_unit="mm")
    # Quantity without specifying unit, must use category's default unit and not raise
    default_when_none = ObtainQuantity(None, "my category")
    assert ObtainQuantity("mm", "my category") == default_when_none


def testDefaultUnitValid(unit_database_custom_conversion) -> None:
    """
    Test the situation where a category does not have the base unit among its valid units
    """
    unit_database = unit_database_custom_conversion
    unit_database.AddCategory("my category", "length", ["mm"])
    assert unit_database.GetCategoryInfo("my category").default_unit == "mm"


def testConvertionWithExponent(unit_database_custom_conversion) -> None:
    """
    Test conversions with different exponents.
    """
    unit_database = unit_database_custom_conversion
    assert approx(abs(100 - unit_database.Convert("length", [("m", 1)], [("cm", 1)], 1)), 5) == 0
    assert approx(abs(10000 - unit_database.Convert("length", [("m", 2)], [("cm", 2)], 1)), 5) == 0

    # Doesn't make sense changing the exponent in the from and to
    with pytest.raises(ValueError):
        unit_database.Convert("length", [("m", 2)], [("m", 1)], 1)


def testAddCategory(unit_database_custom_conversion) -> None:
    """
    Testing if AddCategory gives an error if trying to register twice.
    """
    unit_database = unit_database_custom_conversion
    unit_database.AddCategory(
        "my category", "length", valid_units=["mm", "m", "cm"], default_unit="mm"
    )
    with pytest.raises(UnitsError):
        unit_database.AddCategory("my category", "length", ["mm"], default_unit="mm")

    unit_database.AddCategory("derived category", "length")

    # Create a preferred unit set for categories
    assert unit_database.GetValidUnits("my category") == ["mm", "m", "cm"]
    assert unit_database.GetValidUnits("derived category") == ["m", "mm", "cm", "km"]


def testValidUnits(unit_database_custom_conversion) -> None:
    """
    Test while registering a category that passing None as valid_units returns the
    units of the quantity type.
    """
    unit_database = unit_database_custom_conversion
    unit_database.AddCategory("length", "length", override=True)
    info = unit_database.AddCategory("my category", "length", default_unit="mm")
    length_units = unit_database.GetUnits("length")
    assert info.valid_units is None
    assert unit_database.GetValidUnits("my category") == length_units


def testConvertQuantityTypeCheck(unit_database_custom_conversion) -> None:
    """
    Check for bug when trying to convert a unit to the same unit, but passing
    an invalid quantity type as argument.
    """
    unit_database = unit_database_custom_conversion
    with pytest.raises(InvalidQuantityTypeError):
        unit_database.Convert("XXX", "m", "mm", 100)
    with pytest.raises(InvalidUnitError):
        unit_database.Convert("length", "XXX", "m", 100)
    with pytest.raises(InvalidUnitError):
        unit_database.Convert("length", "m", "XXX", 100)


def testCheckValidUnits(unit_database_custom_conversion) -> None:
    """
    Make sure we check if the given valid_units are actually valid for that quantity_type
    """
    unit_database = unit_database_custom_conversion
    with pytest.raises(ValueError):
        unit_database.AddCategory("my category", "length", default_unit="mm", valid_units="foooo")


def testDiscoverCloseUnitMatches() -> None:
    unit_database = UnitDatabase.CreateDefaultSingleton()
    assert unit_database.FindSimilarUnitMatches("kg/day") == ["kg/d"]
    assert unit_database.FindSimilarUnitMatches("bbls/d") == ["bbl/d", "bbl/d2"]
    assert unit_database.FindSimilarUnitMatches("mg/l") == ["mg/L"]


def testDefaultCaption() -> None:
    unit_database = UnitDatabase.CreateDefaultSingleton()
    category_info = unit_database.GetCategoryInfo("angle per volume")
    assert category_info.caption == "Angle per Volume"


def testAddCategoryBasedOnCategory(unit_database_custom_conversion) -> None:
    unit_database = unit_database_custom_conversion
    unit_database.AddCategory(
        "my category", "length", valid_units=["mm", "m"], default_unit="mm", default_value=0.5
    )

    unit_database.AddCategory("my category 2", from_category="my category", valid_units=["mm"])
    assert unit_database.GetValidUnits("my category 2") == ["mm"]

    unit_database.AddCategory("my category 3", from_category="my category", default_unit="m")
    assert unit_database.GetDefaultUnit("my category 3") == "m"

    unit_database.AddCategory("my category 4", from_category="my category", default_value=0.8)
    assert unit_database.GetDefaultValue("my category 4") == 0.8
    assert unit_database.GetValidUnits("my category 4") == ["mm", "m"]


def testConvertFractionValues(unit_database_custom_conversion) -> None:
    """
    Test unit database handling fraction values.
    """
    db = unit_database_custom_conversion

    # Converting half meter to cm
    value = FractionValue(0, (1, 2))

    value = db.Convert("length", "m", "cm", value)
    assert float(value) == 50


def testRegisterTwoFunctionsForTheSameClass(unit_database_custom_conversion) -> None:
    """
    Test the behavior when we attempt to register two convert functions for the same class
    """

    def ConvertFunction1(*args, **kwargs):
        return 0

    db = unit_database_custom_conversion
    # There is a function specialized in convert fraction values. Attempting to register another
    # function should raise an error

    value = FractionValue(0, (1, 2))
    db.Convert("length", "m", "cm", value)

    with pytest.raises(AssertionError):
        db.RegisterAdditionalConversionType(FractionValue, ConvertFunction1)


def testNumpyConversion(unit_database_custom_conversion) -> None:
    unit_database = unit_database_custom_conversion

    from barril.units.posc import MakeBaseToCustomary, MakeCustomaryToBase

    f_unit_to_base = MakeCustomaryToBase(273.15, 1, 1, 0)
    f_base_to_unit = MakeBaseToCustomary(273.15, 1, 1, 0)
    unit_database.AddUnit(
        "My Temperature",
        "degrees Celsius",
        "degC",
        f_base_to_unit,
        f_unit_to_base,
        default_category=None,
    )
    f_unit_to_base = MakeCustomaryToBase(2298.35, 5, 9, 0)
    f_base_to_unit = MakeBaseToCustomary(2298.35, 5, 9, 0)
    unit_database.AddUnit(
        "My Temperature",
        "degree Fahrenheit",
        "degF",
        f_base_to_unit,
        f_unit_to_base,
        default_category=None,
    )
    f_unit_to_base = MakeCustomaryToBase(0, 1, 1, 0)
    f_base_to_unit = MakeBaseToCustomary(0, 1, 1, 0)
    unit_database.AddUnit(
        "My Temperature", "my Kelvin", "myK", f_base_to_unit, f_unit_to_base, default_category=None
    )

    unit_database.AddCategory(
        "My Temperature",
        quantity_type="My Temperature",
        valid_units=["myK", "degC", "degF"],
        default_unit="myK",
        default_value=2,
        min_value=0,
        max_value=15,
        is_min_exclusive=False,
        is_max_exclusive=False,
        caption="My Temperature",
    )

    import numpy

    values = numpy.array(list(range(10, 13)), numpy.float32)
    arr = Array("My Temperature", values, "degC")
    obtained = arr.GetValues("degF")

    expected = [49.99995041, 51.79994965, 53.5999527]
    assert approx(obtained) == expected


def testNumpyWithMinMax(unit_database_custom_conversion) -> None:
    unit_database = unit_database_custom_conversion
    unit_database.AddCategory(
        "my length",
        "length",
        ["mm", "m"],
        default_unit="m",
        default_value=2,
        min_value=0,
        max_value=15,
        is_min_exclusive=False,
        is_max_exclusive=False,
        caption="My Length",
    )
    import numpy

    values = numpy.array(list(range(10)), numpy.float32)
    Array("my length", values, "mm")

    values = numpy.array(list(range(20000, 20010)), numpy.float32)
    # Raise validation error
    another = Array("my length", values, "mm")
    assert not another.IsValid()
    with pytest.raises(ValueError):
        another.CheckValidity()

    # Don't raise validation error if validate = False.
    arr = Array("my length", values, "mm")
    assert not arr.IsValid()

    another = arr.CreateCopy(values=values, unit="mm")
    assert not another.IsValid()

    # Same checks on FixedArray.
    values = numpy.array(list(range(10)), numpy.float32)
    FixedArray(len(values), "my length", values, "mm")

    values = numpy.array(list(range(20000, 20010)), numpy.float32)
    # Raise validation error
    another = FixedArray(len(values), "my length", values, "mm")
    assert not another.IsValid()

    # Don't raise validation error if validate = False.
    arr = FixedArray(len(values), "my length", values, "mm")
    assert not arr.IsValid()

    another = arr.CreateCopy(values=values, unit="mm")
    assert not another.IsValid()


def testInvalidUnitCategoryDoesntGenerateError(unit_database_custom_conversion) -> None:
    unit_database = unit_database_custom_conversion
    with pytest.raises(ValueError):
        unit_database.AddCategory(
            "my invalid length",
            "length",
            caption="Invalid length",
            # This is not a valid unit
            default_unit="Mm",
            min_value=0,
            max_value=99999.0,
            is_min_exclusive=False,
            is_max_exclusive=False,
            valid_units=["mm", "m"],
        )


def testGetValidUnitsEmptyQuantity(unit_database_custom_conversion) -> None:
    """
    Test that retrieving the valid units for the category of the empty quantity (as created
    by Quantity.CreateEmpty() correctly returns an empty list.
    """
    unit_database = unit_database_custom_conversion
    quantity = Quantity.CreateEmpty()
    assert unit_database.GetValidUnits(quantity.GetCategory()) == []



from typing import Dict
from typing import Tuple

from barril.units.unit_database import InvalidUnitError
from barril.units.unit_database import UnitDatabase


class _SimpleScalar:
    def __init__(self, value, unit, category=None):
        self._value = value
        self._value_quantity = _ObtainQuantity(unit, category)
        self._unit = unit

    def GetValue(self, unit=None):
        return UnitDatabase.GetSingleton().Convert(
            self._value_quantity.GetQuantityType(),
            self._value_quantity.GetUnit(),
            unit or self._unit,
            self._value,
        )

    def SetUnit(self, unit):
        self._unit = unit

    def SetValue(self, value, unit=None):
        self._value_quantity = _ObtainQuantity(
            unit or self._unit, self._value_quantity.GetCategory()
        )
        self._value = value


_quantities_cache: Dict[Tuple[str, str], "_SimpleQuantity"] = {}


def _ObtainQuantity(unit, category=None):
    if not category:
        unit_database = UnitDatabase.GetSingleton()
        category = unit_database.GetDefaultCategory(unit)
        if not category:
            raise InvalidUnitError(unit)

    key = (category, unit)
    quantity = _quantities_cache.get(key)
    if quantity is None:
        quantity = _quantities_cache.setdefault(key, _SimpleQuantity(category, unit))
    return quantity


class _SimpleQuantity:
    def __init__(self, category, unit):
        self._category = category
        self._unit = unit

    def GetCategory(self):
        return self._category

    def GetQuantityType(self):
        unit_database = UnitDatabase.GetSingleton()
        return unit_database.GetCategoryQuantityType(self._category)

    def GetUnit(self):
        return self._unit

    def __str__(self):
        return f"Quantity({self._category}, {self._unit})"


class _ComposingQuantity:
    def __init__(self, category_to_unit_and_exps):
        self._category_to_unit_and_exps = category_to_unit_and_exps


def testScalarRefactor() -> None:
    quantity = _ObtainQuantity("m")
    quantity2 = _ObtainQuantity("m")
    assert "Quantity(length, m)" == str(quantity)
    assert quantity is quantity2

    scalar = _SimpleScalar(10, "m")
    assert 10 == scalar.GetValue("m")
    assert 10 == scalar.GetValue()
    assert 0.01 == scalar.GetValue("km")

    scalar.SetUnit("km")
    assert 10 == scalar.GetValue("m")
    assert 0.01 == scalar.GetValue()
    assert 0.01 == scalar.GetValue("km")

    scalar.SetValue(10)
    assert 10 == scalar.GetValue("km")
    assert 10 == scalar.GetValue()
    assert 10000 == scalar.GetValue("m")



import copy
import math
import traceback
from typing import Any
from typing import Callable
from typing import cast
from typing import Dict
from typing import Hashable
from typing import Iterator
from typing import List
from typing import Optional
from typing import Sequence
from typing import Set
from typing import Tuple
from typing import Type
from typing import TYPE_CHECKING
from typing import TypeVar
from typing import Union

import attr
from oop_ext.foundation.singleton import Singleton

from barril._util.types_ import CheckType
from barril.units.interfaces import UnitExponentTuple


if TYPE_CHECKING:
    from barril.units import Quantity

# Contains the registry for all the available unit types.
__all__ = [
    "CategoryInfo",
    "InvalidOperationError",
    "InvalidQuantityTypeError",
    "InvalidUnitError",
    "UnitDatabase",
    "UnitInfo",
    "UnitsError",
]


_LEGACY_TO_CURRENT: List[Tuple[str, str]] = [
    ("1000ft3", "Mcf"),
    ("1000m3", "Mm3"),
    ("M(ft3)", "MMcf"),
    ("M(m3)", "MMm3"),
    ("k(ft3)", "Mcf"),
    ("Ns/m", "N.s/m"),
]


def FixUnitIfIsLegacy(unit: str) -> Tuple[bool, str]:
    fixed_unit = unit
    try:
        for legacy, current in _LEGACY_TO_CURRENT:
            fixed_unit = fixed_unit.replace(legacy, current)
        return (unit != fixed_unit), fixed_unit
    except:
        return False, unit


class UnitsError(RuntimeError):
    """
    Base class for errors related to units.
    """


class InvalidQuantityTypeError(UnitsError):
    """
    Error raised when an invalid quantity type is found
    """

    def __init__(self, quantity_type: str, available: Optional[List[str]] = None):
        msg = f"Invalid quantity_type: {quantity_type}"
        if available is not None:
            msg += "\nAvailable:\n" + "\n".join(available)
        UnitsError.__init__(self, msg)


class InvalidUnitError(UnitsError):
    """
    Error raised when an invalid unit is found
    """

    def __init__(
        self,
        unit: str,
        quantity_type: Optional[str] = None,
        category: Optional[str] = None,
        valid_units: Optional[List[str]] = None,
    ):
        if quantity_type is not None:
            msg = f"Invalid unit for quantity_type {quantity_type}: {unit}"

        elif category is not None:
            msg = f"Invalid unit for category {category}: {unit}"

        else:
            msg = f"Invalid unit:{unit}"

        if valid_units is not None:
            msg += " [Valid Units: %s]" % valid_units

        UnitsError.__init__(self, msg)


class InvalidOperationError(UnitsError):
    """
    Error raised when some operation but couldn't actually be performed with the given units.
    E.g.: Summing meters + seconds is not valid (while meters + centimeters would be)
    """


class ComposedUnitError(UnitsError):
    """
    Error raised when a composed unit conversion is performed
    E.g.: Velocity_Array * Concentration_Array = m/s.g/cm3
    """


UnaryConversionFunc = Callable[[float], float]


class UnitInfo:
    """
    Holds information about a unit type
    """

    ADD_STR_INFO_TO_UNIT_INFO = False

    def __init__(
        self,
        quantity_type: str,
        name: str,
        unit: str,
        frombase: Union[str, UnaryConversionFunc],
        tobase: Union[str, UnaryConversionFunc],
        default_category: Optional[str] = None,
    ):
        """
        :param name:
            Name of the unit (e.g.: meter, millimiter).

        :param unit:
            String that represents the unit (symbol - e.g.: m, mm).

        :param frombase:
            the formula to convert from the base to this unit.

        :param tobase:
            The formula to convert from the this unit to the base.

        .. note:: frombase or tobase must be defined having the part that should be transformed as
            a %f, %s or just x
        """

        def MakeLambda(s: str) -> UnaryConversionFunc:
            s = s.replace("%s", "x").replace("%f", "x")
            assert "x" in s
            s.replace("x", "float(x)")
            ret = eval("lambda x:%s" % s)
            ret.__has_conversion__ = True
            return ret

        if isinstance(frombase, str):
            frombase_func = MakeLambda(frombase)
        else:
            frombase_func = frombase

        if isinstance(tobase, str):
            tobase_func = MakeLambda(tobase)
        else:
            tobase_func = tobase

        # The functions added to the UnitInfo must say whether they actually have some
        # conversion. If they do, and it's value is True, they may also have
        # __a__, __b__, __c__, __d__ identifying the posc convertions
        # (used for doing conversions for sci20 grid functions in C++)
        # If it's False, that means that the function will have no actual conversion.

        # If not identified, that means that they do have conversions associated
        if not hasattr(tobase_func, "__has_conversion__"):
            tobase_func.__has_conversion__ = True  # type:ignore[attr-defined]

        if not hasattr(frombase_func, "__has_conversion__"):
            frombase_func.__has_conversion__ = True  # type:ignore[attr-defined]

        self.name = name
        self.unit = unit
        self.frombase = frombase_func
        self.tobase = tobase_func
        self.quantity_type = quantity_type
        self.default_category = default_category

        if UnitInfo.ADD_STR_INFO_TO_UNIT_INFO:
            # must be added for the generation of the c++ version of the conversion
            self.frombase_str = frombase
            self.tobase_str = tobase

    def __hash__(self) -> int:
        return hash(self.unit)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(self, UnitInfo):
            return False
        return self.unit == other.unit


@attr.s(auto_attribs=True)
class CategoryInfo:
    """
    Holds information about a category
    """

    category: str = ""
    quantity_type: str = ""
    valid_units: Optional[List[str]] = attr.Factory(list)  # type:ignore[assignment]
    valid_units_set: Set[str] = attr.Factory(set)
    default_unit: Optional[str] = ""
    default_value: float = 0.0
    min_value: Optional[float] = None
    max_value: Optional[float] = None
    is_min_exclusive: bool = False
    is_max_exclusive: bool = False
    caption: str = ""


T = TypeVar("T")
ConversionFunc = Callable[["UnitDatabase", str, str, str, T], T]


class UnitDatabase(Singleton):
    """
    Registry with all the available quantity types and units that represent the physical units.

    Quantity Types represent the type of the unit, for instance length or temperature, as strings.
    Every quantity type has one or more units associated with it.

    Units represent one specific unit inside a quantity_type, for instance meters and centimeters
    for length.
    """

    PREPOSITIONS_IN_CATEGORY_NAME = ("per", "of")

    @classmethod
    def CreateDefaultSingleton(cls) -> "UnitDatabase":
        result = cls(default_singleton=True)
        cls.FillUnitDatabaseWithPosc(result)
        return result

    # Some aliases to some quantity types
    _ADDITIONAL_CATEGORY_ALIASES = {
        "liquid volume flow rate": "volume flow rate",
        "gas volume flow rate": "volume flow rate",
        "liquid volume": "volume",
        "gas volume": "volume",
        "liquid volume per standard volume": "volume per standard volume",
        "gas volume per standard volume": "volume per standard volume",
        "date": "time",
    }

    @classmethod
    def FillUnitDatabaseWithPosc(
        cls, unit_database: "UnitDatabase", fill_categories: bool = True
    ) -> "UnitDatabase":
        """
        Fills a unit database with the posc values.

        :param UnitDatabase unit_database:
            The unit database to be filled.

        :param bool fill_categories:
            Indicates whether for each quantity type a category with the same name should be created.

        :rtype: UnitDatabase
        :returns:
            The unit database passed as a parameter.
        """
        from .posc import FillUnitDatabaseWithPosc

        unit_database.Clear()
        FillUnitDatabaseWithPosc(
            unit_database, fill_categories=fill_categories, override_categories=True
        )

        if fill_categories:
            for quantity_alias, quantity_type in cls._ADDITIONAL_CATEGORY_ALIASES.items():
                unit_database.AddCategory(quantity_alias, quantity_type)

        return unit_database

    def CheckValueForCategory(
        self, category: str, value: float, unit: Optional[str] = None
    ) -> None:
        """
        :param category:
            The category to be checked.

        :param value:
            The value to be checked for the given category.

        :param unit:
            The unit of the value passed (if not available, the default value is considered).
        """
        from ._quantity import ObtainQuantity

        quantity = ObtainQuantity(unit, category)
        quantity.CheckValue(value)

    def CheckDefaultUnitDatabase(self) -> None:
        """
        Checks if this is the default unit-database. If it's not a 'default' unit-database, an
        error is raised together with the stack trace from where it was originally created.

        :raises AssertionError:
            raises error if this was not created as the default unit database.
        """
        if hasattr(self, "_database_created_from"):
            raise AssertionError(
                "Not default unit-database. Creation: \n-------\n%s\n-------\n"
                % (self._database_created_from,)
            )

    # Additional conversions stored at the class (no point in storing them only in an instance,
    # even if it's a singleton). This is done because if a new database is created (say to support just
    # some specific units) it would be painfull to have to register all conversion functions manually.
    # This way we have a singleton that have the application units, other databases can be created
    # but all databases share the same conversion functions.
    #
    # dict of class supported -> conversion function
    # see RegisterAdditionalConversionType
    _additional_conversions: Dict[Type, ConversionFunc] = {}

    def __init__(self, default_singleton: bool = False):
        """
        Initializes the unit manager without any quantity types.

        :param bool default_singleton:
            If True, this is the unit database that's created through the CreateDefaultSingleton
            method. This is needed because any constants created a module or class level must have a
            reference to the default unit-database and not for some other database created in tests
            (so, this flag is used together with CheckDefaultUnitDatabase later on).
        """
        if not default_singleton:
            # If this is not the default singleton, mark from where was it created if we need
            # to check later on.
            from io import StringIO

            s = StringIO()
            traceback.print_stack(file=s)
            self._database_created_from = s.getvalue()

        # Quantities must be cached accordingly to the current unit-database.
        self.quantities_cache: Dict[Hashable, "Quantity"] = {}

        # Dictionary to cache whether a unit is valid in a category.
        self._category_unit_valid: Dict[Tuple[str, str], bool] = {}

        # dict of quantity_type => list of UnitInfo (the first unit in this list is the base unit for
        # the given quantity type)
        self.quantity_types: Dict[str, List[UnitInfo]] = {}

        # dict of unit name => UnitInfo
        self.unit_to_unit_info: Dict[str, UnitInfo] = {}
        self.categories_to_quantity_types: Dict[str, CategoryInfo] = {}

    # ----------------------------------------------- The interfaces below all work with the category

    @classmethod
    def FillSimple(cls, unit_database: "UnitDatabase") -> None:
        unit_database.AddUnitBase("length", "meters", "m")
        unit_database.AddUnit("length", "milimeters", "mm", "%f * 1000.0", "%f / 1000.0")
        unit_database.AddUnit("length", "centimeters", "cm", "%f * 100.0", "%f / 100.0")
        unit_database.AddUnit("length", "kilometers", "km", "%f / 1000.0", "%f * 1000.0")

        unit_database.AddUnitBase("time", "seconds", "s")
        unit_database.AddUnit("time", "minutes", "min", "%f * 60.0", " %f * 60.0")
        unit_database.AddUnit("time", "hours", "h", "%f * 3600.0", " %f * 3600.0")
        unit_database.AddUnit("time", "days", "d", "%f * 86400.0", " %f * 86400.0")

        unit_database.AddCategory("length", "length")
        unit_database.AddCategory("time", "time")

    def AddCategory(
        self,
        category: str,
        quantity_type: Optional[str] = None,
        valid_units: Optional[List[str]] = None,
        override: bool = False,
        default_unit: Optional[str] = None,
        default_value: Optional[float] = None,
        min_value: Optional[float] = None,
        max_value: Optional[float] = None,
        is_min_exclusive: bool = False,
        is_max_exclusive: bool = False,
        caption: str = "",
        from_category: Optional[str] = None,
    ) -> CategoryInfo:
        """Adds a category to the unit-management. If it already exists, throws an error
        if override is not set to True

        :param category:
            The category to be added to the unit-management.

        :param quantity_type:
            The quantity type that this category maps to.

        :param valid_units:
            A set of the valid units for the given category.

        :param override:
            Whether to replace the quantity type for the category.

        :param default_unit:
            The default unit for the category

        :param default_value:
            The default value for the category

        :param min_value:
            Minimum value acceptable. If None, any value can be set.

        :param max_value:
            Maximum value acceptable. If None, any value can be set.

        :param is_min_exclusive:
            If the min_value given is exclusive.

        :param is_max_exclusive:
            If the max_value given is exclusive.

        :param caption:
            User friendly caption for this category.

        :param from_category:
            Category to copy other parameters from.

        :raises UnitsError:
            If the category was already added and override is not set to True
        """
        CheckType(category, str)

        if from_category and quantity_type:
            raise ValueError("cannot pass both quantity_type and from_category")

        if not override and category in self.categories_to_quantity_types:
            raise UnitsError("category %r already registered" % category)

        if min_value is not None and max_value is not None:
            if max_value < min_value:
                raise ValueError(
                    "min_value (%s) must be >= than min_value (%s)" % (min_value, max_value)
                )

        if from_category:
            category_info = self.GetCategoryInfo(from_category)
            quantity_type = category_info.quantity_type
            if valid_units is None:
                valid_units = category_info.valid_units
            if default_unit is None:
                default_unit = category_info.default_unit
            if default_value is None:
                default_value = category_info.default_value
            if min_value is None:
                min_value = category_info.min_value
            if max_value is None:
                max_value = category_info.max_value
            if is_min_exclusive is None:
                is_min_exclusive = category_info.is_min_exclusive
            if is_max_exclusive is None:
                is_max_exclusive = category_info.is_max_exclusive
            if caption is None:
                caption = category_info.caption

        assert quantity_type is not None

        # check if valid_units should inherit from the quantity_type
        if valid_units is not None:
            # valid units given: check if all the given units are valid
            quantity_units = set(self.GetUnits(quantity_type))
            for unit in valid_units:
                if unit not in quantity_units:
                    msg = "unit %r is not valid for quantity type %r.\nQuantity units: %r"
                    raise ValueError(msg % (unit, quantity_type, sorted(quantity_units)))

        # if (min_value is not None or max_value is not None) and default_unit is None:
        if default_unit is None:
            default_unit = self.GetBaseUnit(quantity_type)
            if valid_units and default_unit not in valid_units:
                default_unit = valid_units[0]
        else:
            quantity_units = set(self.GetUnits(quantity_type))
            if default_unit not in quantity_units:
                raise ValueError(
                    "unit %r is not valid for default quantity type %r"
                    % (default_unit, quantity_type)
                )

        # caption
        if not caption:
            caption = category.title()
            # Do not Title prepositions
            for word in self.PREPOSITIONS_IN_CATEGORY_NAME:
                caption = caption.replace(word.title() + " ", word + " ")

        # if the default_value has not been passed
        #   1) min_value defined? ---> min_value
        #   else
        #   2) max_value defined? ---> max_value
        #   else
        #   3) zero
        if default_value is None:
            if is_min_exclusive or is_max_exclusive:
                raise RuntimeError("default_value must be supplied")
            elif min_value is not None:
                default_value = min_value
            elif max_value is not None:
                default_value = max_value
            else:
                default_value = 0.0

        else:  # the default_value is defined
            msg = "Error while adding category %s: default_value %f %s %f"

            if min_value is not None:
                if is_min_exclusive:
                    assert default_value > min_value, msg % (
                        category,
                        default_value,
                        "must be >",
                        min_value,
                    )
                else:
                    assert default_value >= min_value, msg % (
                        category,
                        default_value,
                        "must be >=",
                        min_value,
                    )

            if max_value is not None:
                if is_max_exclusive:
                    assert default_value < max_value, msg % (
                        category,
                        default_value,
                        "must be <",
                        max_value,
                    )
                else:
                    assert default_value <= max_value, msg % (
                        category,
                        default_value,
                        "must be <=",
                        max_value,
                    )

        info = CategoryInfo(
            category=category,
            quantity_type=quantity_type,
            valid_units=valid_units,
            valid_units_set=set(valid_units) if valid_units is not None else set(),
            default_unit=default_unit,
            default_value=default_value,
            min_value=min_value,
            max_value=max_value,
            is_min_exclusive=is_min_exclusive,
            is_max_exclusive=is_max_exclusive,
            caption=caption,
        )

        self.categories_to_quantity_types[category] = info
        return info

    def IsValidCategory(self, category: str) -> bool:
        """
        Check if the given category is valid into the unit database.

        :param category:
            The category to check the validity.

        :returns:
            True is it is a valid category; otherwise False.
        """
        return category in self.categories_to_quantity_types

    def IterCategories(self) -> Iterator[str]:
        """
        Iterator for all the categories.

        :returns:
            An iterator that'll provide all the categories.
        """
        return iter(self.categories_to_quantity_types.keys())

    def GetCategoryInfo(self, category: str) -> CategoryInfo:
        """
        :param category:
            The category we're interested in.

        :returns:
            The category info for the category passed.
        """
        try:
            return self.categories_to_quantity_types[category]
        except KeyError:
            categories_str = ""
            for cat in sorted(self.categories_to_quantity_types.keys()):
                if cat is None:
                    cat = "None"
                categories_str += cat + "\n"
            raise InvalidQuantityTypeError(
                'The category: "%s" is not added to the unit manager.\n'
                "--- Available ---:\n%s" % (category, categories_str)
            )

    def GetCategoryQuantityType(self, category: str) -> str:
        """
        :returns:
            The quantity type of some category.
        """
        return self.GetCategoryInfo(category).quantity_type

    def FindUnitCase(self, category: str, unit: str) -> str:
        """
        Given a unit in any case, returns a unit that is a match with the correct case within
        the unit-database.

        :param category:
            The category for the given unit

        :param unit:
            The unit that should be used to match the case.

        :returns:
            the unit considering the actual case used within the unit database.

        :raises AssertionError:
            if no unit could be found or more than one was found (there should be
            only 1 match considering it in a case-insensitive way).
        """
        category_info = self.GetCategoryInfo(category)
        infos = self.GetInfos(category_info.quantity_type)

        matched = []

        unit_lower = unit.lower()
        for info in infos:
            if info.unit.lower() == unit_lower:
                matched.append(info.unit)

        if len(matched) == 1:
            return matched[0]
        else:
            raise AssertionError(
                "Expected 1 match not considering case for: '%s'. Found: %s"
                % (unit, [u for u in matched])
            )

    def CheckCategoryUnit(self, category: str, unit: str) -> None:
        """
        Check if the given category accepts the passed unit.

        :raises InvalidUnitError:
            if the unit provided is not accepted for this category
        """
        assert category.__class__ == str, f"Expected unit of type str, found {category}"
        assert unit.__class__ == str, f"Expected unit of type str, found {unit}"

        key = (category, unit)
        try:
            # i.e.: if not valid (valid = self._category_unit_valid[key])
            if not self._category_unit_valid[key]:
                raise InvalidUnitError(unit, None, category)
        except KeyError:
            if category.__class__ != str:
                raise TypeError("Only str is accepted. %s is not." % category.__class__)

            try:
                category_info = self.GetCategoryInfo(category)
                # When setting a unit, leave the user set any unit from the quantity type, even if there's
                # a subset for the category (the idea is that the units for the category are only used to
                # filter them in the UI, not really to do internal validations).
                self.CheckQuantityTypeUnit(category_info.quantity_type, unit)
                valid = True
            except UnitsError:
                valid = False

            self._category_unit_valid[key] = valid
            if not valid:
                raise InvalidUnitError(unit, None, category)

    def GetValidUnits(self, category: str) -> List[str]:
        """
        :rtype: list(str)
        :returns:
            The valid units for a given category. If the valid categories weren't given uses the valid units from
            quantity type.
        """
        # Special case: the empty category, as generated by Quantity.CreateEmpty(), has no valid units.
        if category == "":
            return []

        category_info = self.GetCategoryInfo(category)
        if category_info.valid_units is not None:
            return category_info.valid_units
        else:
            if category_info.quantity_type != category:
                return self.GetValidUnits(category_info.quantity_type)

            # the valid units have not been specified for the given category (so, let's return
            # the units for the quantity type)
            return self.GetUnits(category_info.quantity_type)

    def GetDefaultValue(self, category: str) -> float:
        """
        :returns:
            The default value for the given category.
        """
        category_info = self.GetCategoryInfo(category)
        return category_info.default_value

    def GetDefaultUnit(self, category: str) -> str:
        """
        :returns:
            The default unit for the given category.

        .. note::
            This method shouldn't return None, when the default_unit isn't defined for the category
            the quantity type base unit is used.
        """
        category_info = self.GetCategoryInfo(category)
        return category_info.default_unit or ""

    # --------------------- The interfaces below all work with the quantity type and not the category

    def AddUnit(
        self,
        quantity_type: str,
        name: str,
        unit: str,
        frombase: Union[str, UnaryConversionFunc],
        tobase: Union[str, UnaryConversionFunc],
        default_category: Optional[str] = None,
    ) -> None:
        """
        Registers a new unit type.

        :param quantity_type:
            The quantity type for the added unit.

        :param name:
            A user-friendly name for this unit.

        :param unit:
            The unit to be added.

        :param frombase:
            If string, an expression to convert from the base unit of this quantity_type to this
            unit. If callable, must accept a float value that applies the conversion.

        :param tobase:
            If a string, an expression to convert from this unit to the base unit. If callable, must
            accept a float value that applies the conversion.

        .. note:: Each expression must refer to %f or x as the current value of the unit.

        :param default_category:
            The default category for the added unit (if any).
        """
        assert quantity_type is not None
        if unit.__class__ != str:
            raise TypeError("Only str is accepted. %s is not." % unit.__class__)

        if unit is None:
            unit = name
        info = UnitInfo(
            quantity_type, name, unit, frombase, tobase, default_category=default_category
        )
        if unit in self.unit_to_unit_info:
            raise RuntimeError(
                "Unit: %s already added to the unit database for the quantity type: %s (trying to add to: %s)"
                % (unit, self.unit_to_unit_info[unit].quantity_type, quantity_type)
            )
        else:
            self.unit_to_unit_info[unit] = info
        quantity_type_list = self.quantity_types.setdefault(quantity_type, [])

        if unit in [q.unit for q in quantity_type_list]:
            raise RuntimeError("Unit already registered: %s (%s)" % (name, unit))

        quantity_type_list.append(info)

    def AddUnitBase(self, quantity_type: str, name: str, unit: str) -> None:
        """
        Add a base unit type. Inside each quantity_type, there must be at least one Base unit.

        Parameters have the same meaning as in AddUnit().
        """

        def identity(x: Any) -> Any:
            return x

        identity.__has_conversion__ = False  # type:ignore[attr-defined]
        self.AddUnit(quantity_type, name, unit, identity, identity)
        # move the base info to the first position
        # (that's a convention: the base is always in the first position)
        infos = self.quantity_types[quantity_type]
        base = infos[-1]  # was appended to the end in Units.Add
        del infos[-1]
        infos.insert(0, base)

    def GetBaseUnit(self, quantity_type: str) -> Optional[str]:
        """
        :param quantity_type:
            The quantity type for which we want a base unit.

        :returns:
            The base unit of the given quantity_type.

        :raises InvalidQuantityTypeError:
            if the quantity type is not valid
        """
        try:
            infos = self.quantity_types[quantity_type]
            return infos[0].unit
        except KeyError:
            self.CheckQuantityType(quantity_type)
        return None

    def GetDefaultCategory(self, unit: str) -> Optional[str]:
        """
        :param unit:
            The unit for which we want the category.

        :rtype: str or None
        :returns:
            The default category for the passed unit.
        """
        try:
            unit_info = self.unit_to_unit_info[unit]
        except KeyError:
            is_legacy, fixed_unit = FixUnitIfIsLegacy(unit)
            if not is_legacy:
                return None
            unit_info = self.unit_to_unit_info[fixed_unit]
        category = unit_info.default_category
        if category:
            return category
        category = unit_info.quantity_type
        if category in self.categories_to_quantity_types:
            return category
        return None

    def GetQuantityType(self, unit: str) -> Optional[str]:
        """
        :returns:
            A quantity_type that contains the respective unit.
        """
        try:
            return self.unit_to_unit_info[unit].quantity_type
        except KeyError:
            return None

    def FindSimilarUnitMatches(self, unit: str) -> List[str]:
        """
        This function will use heuristics to find similar units in the unit database to the
        passed unit.

        :param unit:
            The unit which doesn't have a direct match in the unit dabatase.

        :returns:
            Returns a list with possible matches for the passed unit, sorted.
        """
        import re

        compiled = re.compile(r"[\./]")
        unit_split = compiled.split(unit.lower())

        close_match = []
        for existing_unit in self.unit_to_unit_info.keys():
            existing_unit_split = compiled.split(existing_unit.lower())
            if len(existing_unit_split) == len(unit_split):
                for a, b in zip(existing_unit_split, unit_split):
                    if not a.startswith(b) and not b.startswith(a):
                        break
                else:
                    close_match.append(existing_unit)

        return sorted(close_match)

    def GetQuantityTypes(self) -> List[str]:
        """
        :returns:
            A list of the available categories, sorted.
        """
        return sorted(self.quantity_types.keys())

    def GetUnits(self, quantity_type: Optional[str] = None) -> List[str]:
        """
        :return:
            The units of that quantity_type (if quantity_type is given) otherwise, returns all
            available units.
        """
        return [x.unit for x in self.GetInfos(quantity_type)]

    def GetUnitName(self, quantity_type: str, unit: str) -> str:
        """
        :returns:
            The user-friendly name for the given unit.
        """
        info = self.GetInfo(quantity_type, unit)
        return info.name

    def GetUnitNames(self, quantity_type: str) -> List[str]:
        """
        :returns:
            The user-friendly names for all the units in the given quantity_type.

        :raises InvalidQuantityTypeError:
        """
        return [x.name for x in self.GetInfos(quantity_type)]

    def GetInfo(
        self, quantity_type: str, unit: str, fix_unknown: bool = False, fix_legacy: bool = True
    ) -> UnitInfo:
        """
        :param bool fix_unknown:
            If True won't raise error if quantity_type is unkwnown (and unit may be anything).
            Returns the unknown unit info in this situation.

        :param bool fix_legacy:
            If True and `unit` is in _LEGACY_TO_CURRENT, then it will return
            the current equivalent unit info.

        :rtype: UnitInfo
        :returns:
            The unit object registered with the given unit

        @raise InvalidQuantityTypeError
        @raise InvalidUnitError
        """

        def TryToGetUnitInfoFromUnit(unit: str) -> Optional[UnitInfo]:
            """
            This is the common case, where the unit matches the quantity type registered.
            """
            try:
                # Common case: unit matches the quantity type registered.
                unit_info = self.unit_to_unit_info[unit]
                if quantity_type == unit_info.quantity_type:
                    return unit_info
            except KeyError:
                pass  # Just ignore and go through the 'uncommon' case.
            return None

        unit_info = TryToGetUnitInfoFromUnit(unit)
        if unit_info is not None:
            return unit_info

        # First check if the quantity_type is a registered category
        try:
            category_info = self.categories_to_quantity_types[quantity_type]
        except KeyError:
            pass
        else:
            quantity_type = category_info.quantity_type

        try:
            quantity_types = self.quantity_types[quantity_type]
        except KeyError:
            raise InvalidQuantityTypeError(quantity_type, sorted(self.quantity_types.keys()))
        else:
            for info in quantity_types:
                if info.unit == unit:
                    return info
            else:
                if fix_unknown:
                    # Before actually triggering the error, handle the unknown case:
                    # We can have an unknown quantity type with a 'known' unit (i.e.: the reader
                    # says it's unknown, but we get a proper label for it in the UI, thus, it's a
                    # known 'unknown' quantity). So, in this case, proceed returning the unknown
                    # quantity unit information.
                    from ._unit_constants import UNKNOWN_QUANTITY_TYPE, UNKNOWN_UNIT

                    if quantity_type == UNKNOWN_QUANTITY_TYPE:
                        quantity_types = self.quantity_types[quantity_type]
                        for info in quantity_types:
                            if info.unit == UNKNOWN_UNIT:
                                return info

                if fix_legacy:
                    is_legacy, fixed_unit = FixUnitIfIsLegacy(unit)
                    if is_legacy:
                        unit_info = TryToGetUnitInfoFromUnit(fixed_unit)
                        if unit_info is not None:
                            return unit_info

                raise InvalidUnitError(
                    unit, quantity_type, valid_units=sorted([info.unit for info in quantity_types])
                )

    def GetInfos(self, quantity_type: Optional[str] = None) -> List[UnitInfo]:
        """
        :returns:
            All UnitInfos from that quantity_type (if quantity_type is given), otherwise return all
            UnitInfos.

        @raise InvalidQuantityTypeError
        """
        if quantity_type is None:
            all_infos = []
            for infos in self.quantity_types.values():
                all_infos.extend(infos)
            return all_infos
        else:
            try:
                return self.quantity_types[quantity_type]
            except KeyError:
                raise InvalidQuantityTypeError(quantity_type, list(self.quantity_types.keys()))

    def CheckQuantityType(self, quantity_type: str) -> None:
        """
        Checks if the quantity type is valid. If it is not, raise an InvalidQuantityTypeError.

        @raise InvalidQuantityTypeError
        @raise InvalidUnitError
        """
        if quantity_type not in self.quantity_types:
            raise InvalidQuantityTypeError(quantity_type, sorted(self.quantity_types))

    def CheckQuantityTypeUnit(self, quantity_type: str, unit: str) -> None:
        """
        Check if the given quantity type has the given unit.

        @raise InvalidUnitError
        """
        # NOTE: Using `fix_legacy=False` because when don't want to fix the legacy unit
        # at this point, we actually want `InvalidUnitError` is the unit is
        # legacy
        self.GetInfo(quantity_type, unit, fix_legacy=False)

    def _ConvertWithExp(
        self,
        quantity_type: str,
        from_unit_exps: Sequence[UnitExponentTuple],
        to_unit_exps: Sequence[UnitExponentTuple],
        value: float,
    ) -> float:
        """
        Converts a value from one unit to another unit considering that the units from and
        to have exponents.

        :param quantity_type:
            The quantity type that has the from and to units.

        :param from_unit_exps:
            The unit and exponent of the "from" unit.

        :param to_unit_exps:
            The unit and exponent of the "to" unit.

        :param float value:
            The value to be converted.

        :rtype: float
        :returns:
            Returns the converted value.
        """
        len_from_unit = len(from_unit_exps)
        len_to_unit = len(to_unit_exps)

        if len_to_unit == 0 or len_from_unit == 0:
            return value

        if len_from_unit != 1:
            raise ComposedUnitError(
                "Can only convert one unit to another (not a composed unit at this point)"
            )

        if len_to_unit != 1:
            raise ComposedUnitError(
                "Can only convert one unit to another (not a composed unit at this point)"
            )

        from_unit, from_exp = from_unit_exps[0]
        to_unit, to_exp = to_unit_exps[0]

        if from_exp != to_exp:
            raise ValueError(
                "Cannot convert among different exponents (%s) to (%s)"
                % ((from_unit, from_exp), (to_unit, to_exp))
            )

        if from_exp == to_exp == 1:
            # Special case handling
            return self.Convert(quantity_type, from_unit, to_unit, value)

        negative = False
        if value < 0.0:
            negative = True
            value = abs(value)

        value = math.pow(value, 1.0 / from_exp)  # Convert from the exponent
        value = self.Convert(quantity_type, from_unit, to_unit, value)
        ret = math.pow(value, to_exp)
        if negative:
            return -ret
        return ret

    @classmethod
    def RegisterAdditionalConversionType(cls, class_: Type, func: ConversionFunc) -> None:
        """
        This function may be used to register conversions for additional classes, not originally
        treated (e.g.: IGridFunction)

        :param type class_:
            This is the class that should be treated. So, when a conversion is requested,
            if a class is an instance of the class passed, it'll be used to do the conversions.

        :param callable func:
            The function that should do the conversion. It'll be called as
            func(unit_database, quantity_type, from_unit_info, to_unit_info, value)
        """
        if class_ not in cls._additional_conversions:
            cls._additional_conversions[class_] = func
        else:
            assert (
                cls._additional_conversions[class_] == func
            ), "The class %s already has a convertion function registered" % (str(class_))

    def Convert(
        self,
        category_or_quantity_type: str,
        from_unit: Union[str, Sequence[UnitExponentTuple]],
        to_unit: Union[str, Sequence[UnitExponentTuple]],
        value: Any,
    ) -> Any:
        """
        Converts a value from one unit to another unit (given the quantity type that contains
        both units), so, note that the quantity type at this point is always the same (can't convert
        units with quantity types that don't match).

        :param category_or_quantity_type:
            The category or quantity type for doing the conversion (if it's a category it's
            converted into the quantity type inside this method).

        :param from_unit:
            A string determining the unit from the value we want to convert or a list of exponents.

        :param to_unit:
            A string determining to convert the value to or a list of exponents.

        :param value:
            The value that should be converted.

        .. note:: that from_unit and to_unit must only have 1 item at this point.

        :returns:
            The converted value
        """
        supported_types = tuple(self._additional_conversions)
        convert_function: Optional[ConversionFunc] = None
        if isinstance(value, supported_types):
            for key, convert_function in self._additional_conversions.items():
                if isinstance(value, key):
                    break  # keep convert_function for later use
            else:
                assert False

        # operations with exponents...
        from_is_list = from_unit.__class__ in (list, tuple)
        to_is_list = to_unit.__class__ in (list, tuple)

        if from_is_list or to_is_list:
            from_unit_exps = cast(list, from_unit) if from_is_list else [(cast(str, from_unit), 1)]
            to_unit_exps = cast(list, to_unit) if to_is_list else [(cast(str, to_unit), 1)]

            # same unit: no conversion needed
            if from_unit_exps == to_unit_exps:
                return value

            if (
                category_or_quantity_type.__class__ in (list, tuple)
                and len(category_or_quantity_type) == 1
            ):
                category_or_quantity_type = category_or_quantity_type[0]
            return self._ConvertWithExp(
                category_or_quantity_type, from_unit_exps, to_unit_exps, value
            )

        from_unit = cast(str, from_unit)
        to_unit = cast(str, to_unit)

        # same unit: no conversion needed
        if from_unit == to_unit:
            return value

        # simple operations (same exponent)
        try:
            quantity_type = self.categories_to_quantity_types[
                category_or_quantity_type
            ].quantity_type
        except KeyError:
            self.CheckQuantityType(category_or_quantity_type)
            quantity_type = category_or_quantity_type

        if convert_function is not None:
            return convert_function(self, quantity_type, from_unit, to_unit, value)

        this = self.GetInfo(quantity_type, from_unit, fix_unknown=True)
        other = self.GetInfo(quantity_type, to_unit, fix_unknown=True)

        if isinstance(value, (float, int)):
            return other.frombase(this.tobase(value))
        else:  # list / tuple
            frombase = other.frombase
            tobase = this.tobase
            values_gen = (frombase(tobase(v)) for v in value)

            if isinstance(value, tuple):
                return tuple(values_gen)
            else:
                return list(values_gen)

    def Clear(self) -> None:
        """
        Removes all the quantity types registered.
        """
        self.quantity_types.clear()
        self.categories_to_quantity_types.clear()
        self.unit_to_unit_info.clear()
        self.quantities_cache.clear()
        self._category_unit_valid.clear()

    # Operations with different quantities ---------------------------------------------------------
    def _DoOperationWithSameQuantity(
        self,
        quantity1: "Quantity",
        quantity2: "Quantity",
        value1: T,
        value2: T,
        operation: Callable[[T, T], T],
    ) -> Tuple["Quantity", T]:
        """
        Given 2 quantities, do an operation that DOES NOT accept the creation of a new composed
        quantity (e.g.: sum, subtraction)
        """
        if quantity1 == quantity2:
            return quantity1, operation(value1, value2)

        else:
            # otherwise, we must transform the units in the quantity1 to their counterparts
            # in the quantity2 (without changing anything in the categories at this time)
            category_to_unit_and_exp1 = copy.deepcopy(quantity1.GetCategoryToUnitAndExps())
            category_to_unit_and_exp2 = copy.deepcopy(quantity2.GetCategoryToUnitAndExps())

            (
                category_to_unit_and_exp1,
                category_to_unit_and_exp2,
                value1,
                value2,
            ) = self._MatchQuantities(
                category_to_unit_and_exp1, category_to_unit_and_exp2, value1, value2
            )

            # at this point the quantity types and units must be the same (the categories may actually
            # be different)
            quantity1 = quantity1.CreateCopyInstance(category_to_unit_and_exp1)
            quantity2 = quantity2.CreateCopyInstance(category_to_unit_and_exp2)

            composing_units1 = set(quantity1.GetComposingUnitsJoiningExponents())
            composing_units2 = set(quantity2.GetComposingUnitsJoiningExponents())
            if composing_units1 != composing_units2:
                if len(composing_units1) == 0:
                    # no unit in the 1st part (just take the 2nd as the correct one)
                    quantity1 = quantity2

                elif len(composing_units2) == 0:
                    pass  # ok, no units in the second part...

                else:
                    raise InvalidOperationError(
                        "Error. Can't do operation because units don't match: (%s != %s)"
                        % (composing_units1, composing_units2)
                    )

            return quantity1, operation(value1, value2)

    def Sum(
        self, quantity1: "Quantity", quantity2: "Quantity", value1: T, value2: T
    ) -> Tuple["Quantity", T]:
        func = lambda a, b: a + b
        return self._DoOperationWithSameQuantity(quantity1, quantity2, value1, value2, func)

    def Subtract(
        self, quantity1: "Quantity", quantity2: "Quantity", value1: T, value2: T
    ) -> Tuple["Quantity", T]:
        func = lambda a, b: a - b
        return self._DoOperationWithSameQuantity(quantity1, quantity2, value1, value2, func)

    def Divide(
        self, quantity1: "Quantity", quantity2: "Quantity", value1: T, value2: T
    ) -> Tuple["Quantity", T]:
        return self._DoOperationResultingInNewQuantity(
            quantity1, quantity2, value1, value2, lambda a, b: a - b, lambda a, b: a / b
        )

    def FloorDivide(
        self, quantity1: "Quantity", quantity2: "Quantity", value1: T, value2: T
    ) -> Tuple["Quantity", T]:
        return self._DoOperationResultingInNewQuantity(
            quantity1, quantity2, value1, value2, lambda a, b: a - b, lambda a, b: a // b
        )

    def Multiply(
        self, quantity1: "Quantity", quantity2: "Quantity", value1: T, value2: T
    ) -> Tuple["Quantity", T]:
        """
        Multiplication with different quantities.

        Rationale:

        - the multiplication of different quantities can result in a derived quantity type.
        - the quantities / values should be transformed if there are quantity types that are
          equal in bot quantities (e.g.: if there's cm in the quantity1 and m in the quantity2,
          the value1 and the quantity1 must be converted to m before proceeding with the
          actual multiplication)
        - note that if a given quantity type has different units in the quantity1 or in
          the quantity2, they must 1st be converted to have the same unit in a given quantity type.
        - after the quantities are compatible, it's just a matter of creating a join of
          the available units and summing the expoents to create the resulting quantity
          (and multiplying the actual values).

        :type quantity1: IQuantity (related to value1)
        :param quantity1:
        :type quantity2: IQuantity (related to value2)
        :param quantity2:
        :type value1: iterable or number
        :param value1:
        :type value2: iterable or number
        :param value2:
        :rtype: tuple(IQuantity, value)
        """
        return self._DoOperationResultingInNewQuantity(
            quantity1, quantity2, value1, value2, lambda a, b: a + b, lambda a, b: a * b
        )

    def _MatchQuantities(
        self,
        category_to_unit_and_exp1: Any,
        category_to_unit_and_exp2: Any,
        value1: Any,
        value2: Any,
    ) -> Tuple[Any, Any, Any, Any]:
        """
        Matches all the units for a given quantity type (so, if a unit 'm' is found, if
        a 'cm' is later found, convert it to 'm' -- as well as it's composing value).

        :rtype: the dicts passed with the quantity types matched to the same units and the corresponding
        values converted to match those changes.
        """
        quantity_types_found_to_used_unit: Dict[Any, Any] = {}

        # 1st thing is putting the same unit for a given quantity type (both sides)
        for c in (category_to_unit_and_exp1, category_to_unit_and_exp2):
            for category, unit_exp in list(c.items()):
                unit, _exp = unit_exp
                quantity_type = self.GetCategoryQuantityType(category)
                used_unit_for_quantity_type = quantity_types_found_to_used_unit.get(quantity_type)
                if used_unit_for_quantity_type is None:
                    quantity_types_found_to_used_unit[quantity_type] = unit
                else:
                    # don't worry about the exponent at this time, just update the unit and the related value.
                    if c is category_to_unit_and_exp1:
                        value1 = self.Convert(
                            quantity_type, unit, used_unit_for_quantity_type, value1
                        )
                    else:
                        value2 = self.Convert(
                            quantity_type, unit, used_unit_for_quantity_type, value2
                        )
                    unit_exp[0] = used_unit_for_quantity_type
        return category_to_unit_and_exp1, category_to_unit_and_exp2, value1, value2

    def _DoOperationResultingInNewQuantity(
        self,
        quantity1: "Quantity",
        quantity2: "Quantity",
        value1: Any,
        value2: Any,
        operation_exp: Any,
        operation: Any,
    ) -> Any:
        """
        Given 2 quantities, do an operation that accepts the creation of a new quantity that is
        composed of the quantities passed (e.g.: division, multiplication).
        """
        from ._quantity import Quantity

        # otherwise, we must transform the units in the quantity1 to their counterparts
        # in the quantity2 (without changing anything in the categories at this time)
        category_to_unit_and_exp1 = quantity1.GetCategoryToUnitAndExpsCopy()
        category_to_unit_and_exp2 = quantity2.GetCategoryToUnitAndExpsCopy()

        (
            category_to_unit_and_exp1,
            category_to_unit_and_exp2,
            value1,
            value2,
        ) = self._MatchQuantities(
            category_to_unit_and_exp1, category_to_unit_and_exp2, value1, value2
        )

        # add the categories to the resulting one
        for category2, (unit2, exp2) in list(category_to_unit_and_exp2.items()):
            if category2 not in category_to_unit_and_exp1:
                exp1 = 0
                category_to_unit_and_exp1[category2] = [unit2, operation_exp(exp1, exp2)]
            else:
                unit_exp1 = category_to_unit_and_exp1[category2]
                unit1, exp1 = unit_exp1  # type:ignore[assignment]
                if unit1 == unit2:
                    unit_exp1[1] = operation_exp(exp1, exp2)  # type:ignore[index]
                else:
                    raise RuntimeError(
                        "This should've been covered already (%s != %s)." % (unit1, unit2)
                    )

        # unit -> expoent
        only_units_expoents: Dict[Any, Any] = {}
        for c, (unit, exp) in list(category_to_unit_and_exp1.items()):
            existing = only_units_expoents.get(unit, 0)
            only_units_expoents[unit] = existing + exp

        # remove the ones that have exponent = 0
        for c, (unit, exp) in list(category_to_unit_and_exp1.items()):
            if exp == 0 or only_units_expoents[unit] == 0:
                # that's ok, it's been removed from the expression...
                del category_to_unit_and_exp1[c]

        result = Quantity.CreateDerived(category_to_unit_and_exp1)
        return result, operation(value1, value2)


class RegisterConversion:
    _registered = False

    @classmethod
    def RegisterNumpyConversion(cls) -> None:
        """
        Register a special unit conversion for numpy arrays.

        :param UnitDatabase db:
            The unit-database instance to register the conversion into.
        """
        import numpy

        def ConvertNumpyArray(
            db: UnitDatabase, quantity_type: str, from_unit: str, to_unit: str, array: Any
        ) -> Any:
            """
            Converts the given numpy array by applying the conversion as if it was a scalar,
            since arrays support the numeric operators, which are applied element-wise.
            """
            from_unit_info = db.GetInfo(quantity_type, from_unit, fix_unknown=True)
            to_unit_info = db.GetInfo(quantity_type, to_unit, fix_unknown=True)

            to_base = from_unit_info.tobase
            from_base = to_unit_info.frombase
            return from_base(to_base(array))

        UnitDatabase.RegisterAdditionalConversionType(numpy.ndarray, ConvertNumpyArray)


RegisterConversion.RegisterNumpyConversion()



from typing import Any
from typing import cast
from typing import Generic
from typing import Optional
from typing import overload
from typing import Tuple
from typing import TYPE_CHECKING
from typing import Union

from ._array import Array
from ._array import ValuesType
from barril.units.unit_database import CategoryInfo
from barril.units.unit_database import UnitDatabase

if TYPE_CHECKING:
    from ._quantity import Quantity
    from barril.units import Scalar

__all__ = ["FixedArray"]


class FixedArray(Array, Generic[ValuesType]):
    """
    Represents an Array with fixed number of elements.


    Like ``Array``, ``FixedArray`` is a ``Generic`` subclass, parametrized by their container type:

    .. code-block:: python

        a = FixedArray(3, [1, 2, 3], "m")
        reveal_type(a)

    .. code-block::

        note: Revealed type is "barril.units._fixedarray.FixedArray[builtins.list*[builtins.int*]]"
    """

    _dimension: Any = None

    @overload
    def __init__(self, dimension: int, category: Union[str, "Quantity"]):
        ...

    @overload
    def __init__(self, dimension: int, values: ValuesType, unit: str):
        ...

    @overload
    def __init__(self, dimension: int, category: str, values: ValuesType, unit: str):
        ...

    @overload
    def __init__(self, dimension: int, category: str, unit: str):
        ...

    @overload
    def __init__(self, dimension: int, category: "Quantity", values: ValuesType):
        ...

    def __init__(  # type:ignore[misc]
        self, dimension: int, category: str, values: Any = None, unit: Any = None
    ) -> None:
        """
        :param int dimension:
            The dimension for this array.

        :type category: string or IQuantity
        :param category:
            The category for the unit point or the IQuantity for this object
            (in this case, the unit will be ignored (if it is passed)).

        :type values: sequence or numpy array
        :param values:
            A sequence with its initial values.

        :param str unit:
            Unit (not used if a quantity is passed).
        """
        if dimension < 2:
            raise ValueError("Dimension MUST be 2 or more")
        self._dimension = dimension

        Array.__init__(self, category, values, unit)

    def _InternalCreateWithQuantity(  # type:ignore[override]
        self,
        quantity: "Quantity",
        values: Optional[ValuesType] = None,
        unit_database: Optional[UnitDatabase] = None,
        dimension: Optional[int] = None,
        value: Optional[ValuesType] = None,
    ) -> None:
        if value is not None:
            if values is not None:
                raise ValueError("Duplicated values parameter given")

            values = value

        assert values is not None
        if dimension is None:
            try:
                if self._dimension is None:
                    self._dimension = len(values)
            except AttributeError:
                pass

            dimension = self._dimension

        elif hasattr(self, "_dimension"):
            if self._dimension is not None and dimension != self._dimension:
                raise ValueError(
                    "Dimension re-definition mismatch: %s != %s" % (self._dimension, dimension)
                )

        if dimension < 2:
            raise ValueError("Dimension MUST be 2 or more")
        self._dimension = dimension

        if values is None:
            values = [0.0] * dimension
        self.CheckValues(values, dimension)

        Array._InternalCreateWithQuantity(self, quantity, values)

    def CreateCopy(  # type:ignore[override]
        self,
        values: Optional[ValuesType] = None,
        unit: Optional[str] = None,
        category: Optional[str] = None,
        **kwargs: object
    ) -> "FixedArray":
        return Array.CreateCopy(
            self, values=values, unit=unit, category=category, dimension=self._dimension, **kwargs
        )

    # Values ---------------------------------------------------------------------------------------
    def _GetDefaultValue(
        self, category_info: CategoryInfo, unit: Optional[str] = None
    ) -> ValuesType:
        return cast(ValuesType, [0.0] * self._dimension)

    def CheckValues(self, values: ValuesType, dimension: Optional[int] = None) -> None:
        """Checks whether the dimensions consistent with the dimensions in this unit point"""
        if dimension is None:
            dimension = self.dimension

        if len(values) != dimension:
            msg = "Values must have %d elements, but has %d"
            raise ValueError(msg % (dimension, len(values)))

    @classmethod
    def CreateEmptyArray(  # type:ignore[override]
        cls, dimension: int, values: Optional[ValuesType] = None
    ) -> "FixedArray":
        """
        Allows the creation of a array that does not have any associated category nor unit.

        :returns:
            Returns an empty array.
        """
        from ._quantity import Quantity

        if values is None:
            values = cast(ValuesType, [0.0] * dimension)

        quantity = Quantity.CreateEmpty()
        return cls.CreateWithQuantity(quantity, dimension=dimension, values=values)

    # Dimension ------------------------------------------------------------------------------------
    def GetDimension(self) -> int:
        return self._dimension

    dimension = property(GetDimension)

    # Equality -------------------------------------------------------------------------------------
    def __eq__(self, other: Any) -> bool:
        return Array.__eq__(self, other) and self.dimension == other.dimension

    def __reduce__(self) -> Any:
        """
        Defining reduce so that we can pickle fixed arrays.
        """
        return (
            FixedArray,
            (self._dimension, self._quantity, self.values, None),  # Unit defined in quantity
        )

    def ChangingIndex(
        self, index: int, value: Union[float, "Scalar", Tuple], use_value_unit: bool = True
    ) -> "FixedArray":
        """
        Creates a FixedArray from based on this FixedArray changing a single value based on the
        passed index.

        i.e.: array.ChangingIndex(0, Scalar(1.0, 'm'))
              will create a new FixedArray where the index == 0 is changed to the passed value.

        :param value:
            The value to be used to set at the given index.

        :param index:
            The index which should be changed.

        :param use_value_unit:
            If True and a Scalar is passed, the newly created array will have the unit of the
            scalar, not of the original array, otherwise (if False) the FixedArray unit will be
            kept.

        :return:
            The created FixedArray.
        """
        from barril.units import Scalar

        if isinstance(value, tuple):
            scalar = Scalar(self.GetValues()[index], self.GetUnit()).CreateCopy(*value)

        elif not isinstance(value, Scalar):
            scalar = Scalar(value, self.GetUnit())

        else:
            scalar = value

        if use_value_unit:
            quantity = scalar.GetQuantity()
        else:
            quantity = self.GetQuantity()

        values = list(self.GetValues(quantity.GetUnit()))
        values[index] = scalar.GetValue(quantity.GetUnit())
        return FixedArray(self.dimension, quantity, tuple(values))

    def IndexAsScalar(self, index: int, quantity: Optional["Quantity"] = None) -> "Scalar":
        """
        :param index:
            The index which should be gotten as a Scalar.

        :param quantity:
            The quantity in which we want the Scalar (uses the FixedArray quantity
            if not passed).

        :return Scalar:
            A Scalar representation of the given index.
        """
        from barril.units import Scalar

        if quantity is None:
            quantity = self.GetQuantity()
        return Scalar(  # type:ignore[call-overload]
            quantity, self.GetValues(unit=quantity.GetUnit())[index]
        )



from typing import Any
from typing import Dict
from typing import Optional

from oop_ext.foundation import callback
from oop_ext.interface import Attribute
from oop_ext.interface import Interface
from oop_ext.interface import TypeCheckingSupport


class IUnitSystem(Interface, TypeCheckingSupport):
    """
    Interface for unit systems. A unit system defines the default unit for all the categories in the
    application (a units mapping).

    :ivar Callback on_default_unit:
        Callback called when a default unit of the unit system changes.
        Signature: (category, unit).
    """

    on_default_unit: callback.Callback2[str, Optional[str]] = Attribute(
        callback.Callback2[str, Optional[str]]
    )

    def __init__(
        self, id: str, caption: str, units_mapping: Dict[str, str], read_only: bool = False
    ) -> None:
        """
        :param id:
            The ID for the unit system.

        :param caption:
            The name of the unit system.

        :param units_mapping:
            A dict that maps each category to a related unit (which will be set as default).
            The valid categories are defined by the coilib50.units.UnitDatabase.

            E.g.:
                units_mapping = {
                    'length' : 'm',
                    'time' : 's',
                    'weight' : 'kg',
                }

        :param read_only:
            Flag that indicates if the unit system will be read only. Default is False.
        """

    def GetCaption(self) -> str:  # type:ignore[empty-body]
        """
        :returns:
            Returns a user-friendly caption.
        """

    def SetCaption(self, caption: str) -> None:  # type:ignore[empty-body]
        """
        Sets the caption of the unit system.

        :param caption:
            The caption.
        """

    def GetId(self) -> Optional[str]:  # type:ignore[empty-body]
        """
        :returns:
            Returns the ID of the unit system.
        """
        ...

    def GetUnitsMapping(self) -> Dict[str, str]:  # type:ignore[empty-body]
        """
        :returns:
            Returns the units mapping set of the unit system.
        """
        ...

    def GetDefaultUnit(self, category: str) -> Optional[str]:  # type:ignore[empty-body]
        """
        :returns:
            Gets the default unit for the given category. If it returns None, it should not change
            objects of that category.
        """
        ...

    def SetDefaultUnit(self, category: str, unit: str) -> None:  # type:ignore[empty-body]
        """
        Changes the default unit for the given category.

        :param unicode category:
            The category which will have the default unit changed.

        :param unicode unit:
            The new default unit.
        """
        ...

    def IsReadOnly(self) -> bool:  # type:ignore[empty-body]
        """
        Retrieves if the unit system is read-only or not.

        :returns:
            True or False indicating the read-only property of the unit system.
        """

    def SetReadOnly(self, read_only: bool) -> None:
        """
        Changes the read-only status of this unit system.

        :param bool read_only:
            Flag indicating if the unit system should be read-only.
        """

    def RemoveCategory(self, category: str) -> None:
        """
        Removes the given category from the categories map.

        :param category:
            The category to remove
        """

    def __eq__(self, other: Any) -> bool:  # type:ignore[empty-body]
        """eq operator"""



"""
Barril basic interfaces are defined here.

Basic concepts used in the interfaces:

Category: Identifies the category for some value. This is defined in the application level. E.g.:
    Well length, Well diameter.

QuantityType: The type of the quantity we're interested in. E.g.: length, depth (this was previously
    defined as Measure, but with different meanings among applications, so, the name Measure has been
    'deprecated' in the favor of this one, so that old-habits about it are not continued -- and it is also
    referenced in posc).

Unit: the unit itself. E.g.: m, m/s, kg

Note: The naming conventions were gathered from posc:
    http://www.posc.org/ebiz/pefxml/patternsobjects.html http://www.posc.org/refs/poscUnits20.xml
"""
from typing import Any
from typing import Dict
from typing import Generic
from typing import Iterator
from typing import List
from typing import Optional
from typing import overload
from typing import Tuple
from typing import TYPE_CHECKING
from typing import TypeVar
from typing import Union

from oop_ext.interface import Interface
from oop_ext.interface import TypeCheckingSupport
from typing_extensions import Protocol

if TYPE_CHECKING:
    from barril.units import UnitDatabase

__all__ = [
    "IQuantity",
    "IScalar",
    "IQuantity2",
    "IQuantity3",
    "IQuantity6",
    "IObjectWithQuantity",
    "IArray",
    "ValuesType",
]


class IQuantity(Interface, TypeCheckingSupport):
    """
    The quantity is an object that has its associated category, quantity type and unit.

    It is important that each value in the application has an associated quantity (because
    otherwise, a value may be meaningless).
    """

    def GetCategory(self) -> str:  # type:ignore[empty-body]
        """
        :returns:
            The constant category for this quantity.
        """

    def GetQuantityType(self) -> str:  # type:ignore[empty-body]
        """
        :returns:
            The constant name of the quantity type for this quantity.
            This method may be slow.
        """

    def GetUnit(self) -> str:  # type:ignore[empty-body]
        """
        :returns:
            The unit for this quantity.
        """


UnitExponentTuple = Tuple[str, int]


class IQuantity2(Interface, TypeCheckingSupport):
    """
    Optional interface to the IQuantity used to deal with operations dealing with different
    units which result in derived units.

    With that in mind, this interface defines that all the information (category, quantity type and
    unit) may be dicts instead of only strings, where a dict is a string that points to a given
    expoent.

    In this way, a unit defined as m2/s would be a dict: {'m':2, 's':-1} that is later used to
    do operations.

    This return type is shared for all the 'Gets' defined in this interface

    Quantities that do not define this interface may not be used in operations that may result in
    a derived quantity.
    """

    def GetComposingCategories(self) -> Union[Tuple[str, ...], str]:  # type:ignore[empty-body]
        """
        :returns:
            A tuple with the categories used.

            A special use-case is provided if there's a single category with an exponent 1, in which
            case a string is returned (optimization).
            I.e.: instead of returning tuple('length',), 'length' will be returned.

        .. see:: GetComposingUnits to return the actual units/exponents for the categories
        """

    def GetComposingUnits(  # type:ignore[empty-body]
        self,
    ) -> Union[Tuple[UnitExponentTuple, ...], str]:
        """
        :rtype: tuple(tuple(str, int)) or str
        :returns:
            A list that will have an entry for each category in this quantity.
            e.g.:tuple(('m',1), ('m',1))

            A special use-case is provided if there's a single unit with an exponent 1, in which
            case a string is returned (optimization).
            I.e.: instead of returning tuple(('m', 1)), 'm' will be returned.

        .. see:: GetComposingUnitsJoiningExponents
        """

    def GetComposingUnitsJoiningExponents(  # type:ignore[empty-body]
        self,
    ) -> Tuple[UnitExponentTuple, ...]:
        """
        :rtype: tuple(tuple(str, int))
        :returns:
            A list with an entry pointing to the total number of exponents found for each unit.
            e.g.:tuple(('m',2))

        .. see:: GetComposingUnits
        """

    def GetCategoryToUnitAndExps(  # type:ignore[empty-body]
        self,
    ) -> Dict[str, List[UnitExponentTuple]]:
        """
        Return an ordered dictionary with the name of a category -> list with 2 elements:
        [unit, exp] that determines the information about categories, quantities and their
        relations to an exponent.

        .. note:: The same INTERNAL REFERENCE should be returned, and not a copy (so, clients that
        change it WILL cause side-effects in the internal dict -- so, creating a deepcopy is
        the clients responsibility)
        """


class IQuantity3(Interface, TypeCheckingSupport):
    def GetUnitDatabase(self) -> "UnitDatabase":  # type:ignore[empty-body]
        """
        :returns:
            The UnitDatabase to which this quantity is associated.
        """


class IQuantity6(Interface, TypeCheckingSupport):
    """
    Interface that defines a way to get the unit caption properly. This means that the
    translation will be applied (without changing the internal unit, just its representation
    for the user) and additional info may be set for when an unknown unit is available.
    """

    def GetUnitCaption(self) -> str:  # type:ignore[empty-body]
        """
        :returns:
            The text related to this quantity that should be shown to the user.
        """

    def SetUnknownCaption(self, caption: str) -> None:
        """
        :param str caption:
            The caption to be shown to the user when it's an unknown unit.

            An empty caption means that the regular unit should be shown to the user even
            when unknown.
        """

    def GetUnknownCaption(self) -> str:  # type:ignore[empty-body]
        """
        :rtype: str
        :returns:
            The caption that's set to be shown to the user when it's an unknown unit.
        """


class IObjectWithQuantity(Interface, TypeCheckingSupport):
    """
    Interface provided for an object that has an associated quantity.
    """

    def GetQuantity(self) -> IQuantity:  # type:ignore[empty-body]
        """
        :returns:
            The quantity that is associated with this object.
            The quantity may be writable or read-only, depending on the object.
            Actually, it may also implement some of the other IQuantity* interfaces.
        """


class IScalar(IObjectWithQuantity, IQuantity):
    """
    Defines a value with an associated unit, without any support for modification, validation or
    unit conversion.
    """

    def GetValue(self, unit: Optional[str] = None) -> float:  # type:ignore[empty-body]
        """
        :param unit:
            The unit we want to get the value back.

        :returns:
            The value stored in this scalar.
        """

    def GetValueAndUnit(self) -> Tuple[float, str]:  # type:ignore[empty-body]
        """
        :returns:
            Tuple with value and current unit name.
        """

    def IsValid(self) -> bool:  # type:ignore[empty-body]
        """
        :returns:
            True if the current value is valid, False otherwise.
        """


class IArray(IObjectWithQuantity, TypeCheckingSupport):
    """
    The Array defines a list of values with a quantity (so, it implements the IObjectWithQuantity
    interface)
    """

    def GetValues(self, unit: Optional[str] = None) -> "MinimalSequence":  # type:ignore[empty-body]
        """
        :param unit:
            This is the unit in which we want the value.

        :returns:
            The values of this array.
        """


ValueType = TypeVar("ValueType", covariant=True)


ValuesType = TypeVar("ValuesType", bound="MinimalSequence")


class MinimalSequence(Protocol[ValueType]):
    """
    Protocol containing the minimum methods an object must implement to be
    considered a Sequence for IArray.

    We cannot use ``typing.Sequence`` as that defines many methods which
    ``numpy`` doesn't implement.
    """

    def __iter__(self) -> Iterator[ValueType]:
        ...

    def __len__(self) -> int:
        ...

    @overload
    def __getitem__(self, item: int) -> ValueType:
        ...

    @overload
    def __getitem__(self, item: slice) -> "MinimalSequence[ValueType]":
        ...

    def __getitem__(self, item: Any) -> Any:
        ...



import locale
import math
import re
from typing import Any
from typing import Callable
from typing import Optional
from typing import Tuple
from typing import Union

from barril._util.types_ import CheckType
from barril.basic.format_float import FloatFromString
from barril.basic.fraction import Fraction


class FractionValue:
    """
    Simple class that acts as a container for both a number and a fraction. Useful
    to aggregate both concepts in a single object.

    To get the actual value as a float, just call float() on the FractionValue object.

    Usage example:
        foo = FractionValue(1.25, Fraction(3, 4))
        print float(foo) # 2.0 == 1.25 + 3/4
        print str(foo) # '1.25 3/4'

    .. note:: neither number nor fractional part can be None, TypeError will be raised whenever an
           attempt to set any of these parts as None is performed.
    .. note:: there's no unit associated with this object. Check L{FractionScalar} for that.
    .. note:: no arithmetic operators have been implemented, but they're certainly interesting if the
           need arises.
    """

    def __init__(
        self, number: float = 0.0, fraction: Union[Fraction, Tuple[float, float]] = (0.0, 1.0)
    ) -> None:
        """
        Constructor.

        :param number:
            The number part for the FractionValue

        :param fraction:
            The fraction for this FractionValue object. A pair (numerator, denominator)
            is also accepted.

        .. note:: Fraction default in init method is a tuple to avoid the caveat of sharing same
            default fraction instance among different instances of this class.
        """
        self.SetNumber(number)
        self.SetFraction(fraction)

    # Number ---------------------------------------------------------------------------------------

    def SetNumber(self, number: float) -> None:
        """
        Sets the number part.

        :param number:
            The integer part.
        """
        CheckType(number, (int, float))
        self._number = number

    def GetNumber(self) -> float:
        """
        :returns:
            Returns the number part.
        """
        return self._number

    number = property(GetNumber, SetNumber)

    # Fraction -------------------------------------------------------------------------------------

    def SetFraction(self, fraction: Union[Fraction, Tuple[float, float]]) -> None:
        """
        Sets the fractional part of this object.

        :param fraction:
            The fraction for this FractionValue object. A pair (numerator, denominator)
            is also accepted.
        """
        CheckType(fraction, (Fraction, tuple))

        # convert a tuple
        if isinstance(fraction, tuple):
            if len(fraction) != 2:
                raise ValueError(f"Expected a tuple (numerator, denominator), got {fraction!r}")
            fraction = Fraction(*fraction)

        self._fraction = fraction

    def GetFraction(self) -> Fraction:
        """
        Returns the fractional part for this object.

        :rtype: Fraction
        :returns:
            The fractional part.
        """
        return self._fraction

    fraction = property(GetFraction, SetFraction)

    # Str/Repr -------------------------------------------------------------------------------------

    def GetLocalizedString(self) -> str:
        """
        :returns:
            Returns a locale-dependent and user-friendly string representation.
        """
        from barril.basic.format_float import FormatFloat

        return self.__FormatToString(FormatFloat)  # type:ignore[arg-type]

    def GetLocalizedFraction(self) -> str:
        """
        :returns:
            Returns a locale-dependent and user-friendly string representation only of fraction
            part.
        """
        from barril.basic.format_float import FormatFloat

        return self.__FormatFractionToString(FormatFloat)  # type:ignore[arg-type]

    def __str__(self) -> str:
        """
        :returns:
            Returns a locale-agnostic and user-friendly string representation.
        """
        return self.__FormatToString(lambda pattern, value: pattern % value)

    def __FormatToString(self, formatter: Callable[[str, Union[Fraction, float]], str]) -> str:
        """
        :param formatter:
            A function that receives, respectively, string format to convert value and a float
            value. Must return a string representation of value.

        :returns:
            Fraction values follow format "%(number)g %(fraction)s".
        """
        formatted = formatter("%g", self._number)
        if float(self._fraction) != 0.0:
            formatted = "{} {}".format(formatted, self.__FormatFractionToString(formatter))
        return formatted

    def __FormatFractionToString(
        self, formatter: Callable[[str, Union[Fraction, float]], str]
    ) -> str:
        """
        :param callable formatter:
            A function that receives, respectively, string format to convert value and a float
            value. Must return a string representation of fractional part.

        :returns:
            Fractional part follows format "%(fraction)s". Whenever fractional part is equivalent to
            zero it is omitted from string to be more user-friendly.
        """
        formatted = ""
        if float(self._fraction) != 0.0:
            formatted = formatter("%s", self._fraction)
        return formatted

    def __repr__(self) -> str:
        """
        :returns:
            Returns the programmer-friendly string representation.
        """
        return f"FractionValue({self._number:g}, {self._fraction})"

    # Equality -------------------------------------------------------------------------------------

    def __eq__(self, other: Any) -> bool:
        """
        Implements '==' binary operator.

        :param FractionValue other:
            Other FractionValue object
        """
        if type(self) is not type(other):
            return False
        return self._number == other._number and self._fraction == other._fraction

    def __ne__(self, other: Any) -> bool:
        """
        Implements '!=' binary operator.

        :param FractionValue other:
            Other FractionValue object
        """
        return not self == other

    # Comparisons ----------------------------------------------------------------------------------

    def __lt__(self, other: Any) -> bool:
        """
        Implements '<' binary operator.

        :param FractionValue other:
            Other FractionValue object
        """
        return float(self) < float(other)

    def __le__(self, other: Any) -> bool:
        """
        Implements '<=' binary operator.

        :param FractionValue other:
            Other FractionValue object
        """
        return float(self) <= float(other)

    def __gt__(self, other: Any) -> bool:
        """
        Implements '>' binary operator.

        :param FractionValue other:
            Other FractionValue object
        """
        return float(self) > float(other)

    def __ge__(self, other: Any) -> bool:
        """
        Implements '>' binary operator.

        :param FractionValue other:
            Other FractionValue object
        """
        return float(self) >= float(other)

    # Float ----------------------------------------------------------------------------------------

    def __float__(self) -> float:
        """
        Implements float() operator.

        :returns:
            The float value of this FractionValue
        """
        return self._number + float(self._fraction)

    # Copy -----------------------------------------------------------------------------------------

    def __copy__(self) -> "FractionValue":
        """
        Returns a copy of this fraction value object.

        :rtype: FractionValue
        :returns:
            The copy.
        """
        return self.__class__(self._number, (self._fraction.numerator, self._fraction.denominator))

    # FromString -----------------------------------------------------------------------------------

    # : regular expression to validate input text as a FractionValue, accepting:
    # : "5.0" (single number, without fraction)
    # : "5.0 3/4" (number followed by a fraction)
    # : "3/4" (single fraction)
    NUMBER_EXPR = r"""
        [-+]?\d+(?:(\.|\,)\d+)     # number, float, accepting "." or ",".
    """

    FRACTION_PART_EXPR = (
        r"""
        (?:
        (?P<numerator>%s?)         # numerator
        \s*/\s*                    # '/' no matter if exist spaces
        (?P<denominator>\d+)       # only integers
        )?
    """
        % NUMBER_EXPR
    )

    FRACTION_COMPLETE_EXPR = r"""
        (?P<float>%s?)             # The whole number
        \s*                        # space between number and fraction
        (%s)?                      # entire fractional part is optional
        """ % (
        NUMBER_EXPR,
        FRACTION_PART_EXPR,
    )

    _FRACTION_RE = re.compile(
        "^" + FRACTION_COMPLETE_EXPR + "$",
        re.VERBOSE,  # Enables the documentation in the regular expression.
    )

    _FRACTION_PARTIAL_RE = re.compile(
        "^" + FRACTION_PART_EXPR + "$",
        re.VERBOSE,  # Enables the documentation in the regular expression.
    )

    @classmethod
    def CreateFromString(cls, text: str, consider_locale: bool = True) -> "FractionValue":
        """
        Create a FractionValue from a string.

        :param text:
            The text with a fraction-value, that is, three integers separated by space and slash
            respectively.
            <value> <value>/<value>

        :param consider_locale:
            Consider the locale on converting string to float

        :returns:
            The fraction value

        :raises ValueError:
            If the text is not convertible to a FractionValue
        """
        text = str(text).strip()

        string_to_float_callable: Callable[[str], float]
        if consider_locale:
            string_to_float_callable = FloatFromString
        else:
            string_to_float_callable = float

        # First try to match only the fractional part.
        m = cls._FRACTION_PARTIAL_RE.match(text)
        number_match: Optional[str]
        if m is not None:
            number_match = None
        else:
            # If can't match a fraction try a mixed number
            m = cls._FRACTION_RE.match(text)
            if m is None:
                raise ValueError('Please enter a text in the form: "5 3/4"')
            number_match = m.group("float")

        if number_match is None:
            number = 0.0
        else:
            number = string_to_float_callable(number_match)

        if m.group("numerator") is not None and m.group("denominator") is not None:
            numerator = string_to_float_callable(m.group("numerator"))
            denominator = locale.atoi(m.group("denominator"))
            fraction = Fraction(numerator, denominator)
        else:
            fraction = Fraction(0.0, 1.0)

        return FractionValue(number, fraction)

    @classmethod
    def MatchFractionPart(cls, text: str) -> None:
        """
        Verify if the given text matchs only the fractional part of a fraction value.

        :param text:
            The given text to verify matching

        @raise: ValueError
            When the text not match the expression, like "3 1/2"
        """
        text = str(text).strip()
        m = cls._FRACTION_PARTIAL_RE.match(text)
        if m is None:
            raise ValueError('Please enter a text in the form: "3/4"')

    @classmethod
    def CreateFromFloat(cls, value: float) -> "FractionValue":
        """
        Convert a float or integer value to their representation in fraction.

        Example:
            result = FractionValue.CreateFromFloat(0.375) // returns FractionValue(fraction=(3,8))

        :param value:
            The value that will be converted to fraction.

        :returns:
            The fractional number.

        .. note:: This code was copied from Alan Hensel at http://www.mindspring.com/~alanh/fracs.html
        and changed to work with negative inputs.
        """

        def FindNumerator(
            number_of_digits_past_decimal: int, dividend: float, divisor: float
        ) -> float:
            result = dividend
            while number_of_digits_past_decimal > 0 and result % divisor == 0:
                result /= divisor
                number_of_digits_past_decimal -= 1
            return result

        def GetMaxNumerator(value: float) -> int:
            str_value = str(value)
            ixe = str_value.lower().find("e")

            # How many digits?
            if ixe == -1:
                f2 = str_value
            else:
                f2 = str_value[0:ixe]
            ix = f2.find(".")
            if ix == -1:
                digits = f2
            elif ix == 0:
                digits = f2[1:]
            elif ix < len(f2):
                digits = f2[0:ix] + f2[ix + 1 :]
            else:
                digits = ""
            number_of_digits = len(digits)

            # How many integer digits?
            if value == 0.0:
                number_of_integer_digits = 0
            else:
                number_of_integer_digits = len(str_value)

            # How many digits past decimal?
            number_of_digits_past_decimal = number_of_digits - number_of_integer_digits

            # Find the numerator when the divisor is 2
            result = FindNumerator(number_of_digits_past_decimal, dividend=float(digits), divisor=2)

            # Find the numerator when the divisor is 5
            result = FindNumerator(number_of_digits_past_decimal, dividend=result, divisor=5)

            return int(result)

        def GetNaN() -> float:
            infinity = 1e10000
            return infinity / infinity

        def GetFractionalPart(value: float) -> float:
            str_value = str(value)
            pos = str_value.find(".")
            return float("0." + str_value[pos + 1 :])

        # The value is None?
        if value is None:
            return None

        # The type is float or integer?
        value_type = type(value)
        if value_type != float and value_type != int:
            raise TypeError("The value type must be int or float")

        # The fractional part exist?
        if float(value) % 1 == 0.0:
            return FractionValue(value)

        # Keep the sign of the given value and work with its absolute value. The sign will be added
        # to the generated fraction at the end.
        sign = value / abs(value)
        value = abs(value)

        integer_part = int(value)
        fractional_part = GetFractionalPart(value)
        str_fractional_part = str(fractional_part)
        max_numerator = GetMaxNumerator(fractional_part)

        # Hold the calculation
        calculation = GetNaN()
        previous_calculation = GetNaN()

        # Numerators and denominators
        numerator = 0
        denominator = 0
        numerators = [0, 1]
        denominators = [1, 0]

        i = 2
        while i < 1000:
            L2 = int(math.floor(fractional_part))

            numerators.append(L2 * numerators[i - 1] + numerators[i - 2])
            if abs(numerators[i]) > max_numerator:
                break

            denominators.append(L2 * denominators[i - 1] + denominators[i - 2])

            calculation = numerators[i] / float(denominators[i])
            if str(calculation) == str(previous_calculation):
                break

            numerator = abs(numerators[i])
            denominator = abs(denominators[i])

            if str(calculation) == str_fractional_part:
                break

            previous_calculation = calculation

            fractional_part = 1 / float(fractional_part - L2)

            i += 1

        if str(numerator) == str(denominator):
            return FractionValue(integer_part)

        return FractionValue(sign * integer_part, (sign * numerator, denominator))



from typing import Any

import pytest

from barril.units.unit_system import UnitSystem


@pytest.fixture
def units_mapping_1():
    units_mapping = {"length": "m", "viscosity": "cP"}
    return units_mapping


@pytest.fixture
def units_mapping_2():
    units_mapping = {"length": "km", "viscosity": "cP"}
    return units_mapping


def testUnitSystemManager(units_mapping_1) -> None:
    system1 = UnitSystem("system1", "My System", units_mapping_1)
    system1.SetReadOnly(True)
    assert system1.GetId() == "system1"
    assert system1.GetCaption() == "My System"
    assert system1.IsReadOnly()
    assert system1.GetDefaultUnit("length") == "m"
    assert system1.GetDefaultUnit("viscosity") == "cP"
    assert system1.GetDefaultUnit("pressure") is None
    assert system1.GetDefaultUnit("") is None


def testEquality(units_mapping_1, units_mapping_2) -> None:
    system1 = UnitSystem("system1", "My System", units_mapping_1, True)
    system2 = UnitSystem("system1", "My System", units_mapping_1, True)
    system3 = UnitSystem("system3", "My System 3", units_mapping_2, False)

    assert system1 == system2
    assert not system1 != system2

    assert system2 != system3
    assert not system1 == system3

    assert system1 is not None


def testSetDefaultUnitCallback(units_mapping_1) -> None:
    """
    Implements on_default_unit on unit system, called when the default unit changes.
    """

    class MockClass:
        callback_params: Any = None

        def OnDefaultUnit(self, category, unit):
            self.callback_params = (category, unit)

    obj = MockClass()
    system1 = UnitSystem("system1", "My System", units_mapping_1)
    system1.on_default_unit.Register(obj.OnDefaultUnit)

    system1.SetDefaultUnit("length", "km")
    assert obj.callback_params == ("length", "km")

    system1.SetDefaultUnit("viscosity", "P")
    assert obj.callback_params == ("viscosity", "P")


def testRemoveCategory(units_mapping_1) -> None:
    class MockClass:
        update_notified: Any = None

        def CallbackUnit(self, category, new_unit):
            if category == "length":
                self.update_notified = True

    system_1 = UnitSystem("system1", "My System", units_mapping_1)

    obj = MockClass()
    system_1.on_default_unit.Register(obj.CallbackUnit)
    assert system_1.GetDefaultUnit("length") == "m"

    # Removing the given quantity
    obj.update_notified = False
    system_1.RemoveCategory("length")
    assert system_1.GetDefaultUnit("length") is None

    assert obj.update_notified



import copy

import pytest

from barril import units
from barril.basic.fraction import FractionValue


def testFractionScalar() -> None:
    # create our scalar
    f = units.FractionScalar("length", value=FractionValue(250, (3, 4)), unit="m")

    # check formatting
    assert f.GetFormatted() == "250 3/4 [m]"

    # check conversion
    # f.unit = 'km'
    assert f.GetValue("km") == FractionValue(0.25, (3 / 1000.0, 4))
    assert f.GetValue("m") == FractionValue(250.0, (3, 4))

    # test no fraction part
    f = f.CreateCopy(value=FractionValue(0.25), unit="km")
    assert f.value == FractionValue(0.25)
    assert f.GetFormatted() == "0.25 [km]"

    # set fraction again
    f = f.CreateCopy(value=FractionValue(250, (3, 4)), unit="m")
    assert f.GetValue("km") == FractionValue(0.25, (3 / 1000.0, 4))
    assert f.GetFormatted() == "250 3/4 [m]"
    assert f.GetValue("m") == FractionValue(250, (3, 4))

    f = f.CreateCopy(value=FractionValue(0.25, (3, 4)), unit="km")
    assert f.value == FractionValue(0.25, (3, 4))
    assert f.GetFormatted() == "0.25 3/4 [km]"
    assert f.GetValue("m") == FractionValue(250.0, (3000, 4))

    with pytest.raises(AttributeError):
        setattr(f, "value", 10)


def testCopy() -> None:
    f = units.FractionScalar("length", value=FractionValue(250.0, (3, 4)), unit="m")
    c = copy.copy(f)
    assert c.value == FractionValue(250.0, (3, 4))


def testEquality() -> None:
    def Create(number, fraction):
        return units.FractionScalar("length", value=FractionValue(number, fraction), unit="m")

    assert Create(250, (3, 4)) == Create(250, (3, 4))
    assert not Create(250, (3, 4)) != Create(250, (3, 4))

    assert Create(100, (3, 4)) != Create(250, (3, 4))
    assert not Create(100, (3, 4)) == Create(250, (3, 4))

    a = units.FractionScalar("length", value=1.0, unit="m")
    b = units.FractionScalar("length", value=2.0, unit="m")
    c = units.FractionScalar("length", value=1.0, unit="in")
    d = units.FractionScalar("depth", value=1.0, unit="m")
    e = None
    f = units.FractionScalar("length", value=1.0, unit="m")

    assert a != b
    assert a != c
    assert a != e
    assert a != d

    assert a == f


def testRepr() -> None:
    assert repr(FractionValue(250, (3, 4))) == "FractionValue(250, 3/4)"


def testFormatValue() -> None:
    f = units.FractionScalar("length", value=FractionValue(250.0, (3, 4)), unit="m")
    assert f.GetFormattedValue() == "250 3/4"


def testComparison() -> None:
    f1 = units.FractionScalar("length", value=FractionValue(10), unit="in")
    f3 = units.FractionScalar("volume", value=FractionValue(4), unit="m3")
    with pytest.raises(TypeError):
        f1 < f3

    f2 = units.FractionScalar("length", unit="in")
    f1 = f1.CreateCopy(value=FractionValue(250, (1, 2)), unit="m")
    f2 = f2.CreateCopy(value=FractionValue(220, (3, 4)), unit="m")

    assert f1 > f2
    assert f2 < f1


def testFractionScalarConversion() -> None:
    db = units.UnitDatabase()
    db.AddUnit("length", "milimeters", "mm", "%f * 1000.0", "%f / 1000.0")
    db.AddUnitBase("length", "meters", "m")

    f = units.FractionScalar("length", value=FractionValue(3, (1, 2)), unit="m")

    converted = db.Convert("length", "m", "mm", f.value)
    assert converted == FractionValue(3500)


def testFractionScalarInvalidValue(unit_database_len) -> None:
    db = unit_database_len

    db.AddCategory("another-length", "length", min_value=5, max_value=15)

    scalar = units.FractionScalar("another-length", value=FractionValue(1, (1, 5)), unit="m")
    assert not scalar.IsValid()
    with pytest.raises(ValueError):
        scalar.CheckValidity()

    # By default the validation will be performed, 10 is a valid value
    scalar = scalar.CreateCopy(value=FractionValue(10))
    assert scalar.IsValid()

    # Even invalid ,the scalar returns the value, unit and a formatted text.
    another = units.FractionScalar("another-length", value=FractionValue(3000), unit="m")
    assert not another.IsValid()
    assert another.GetValue("m") == FractionValue(3000)
    assert another.GetUnit() == "m"
    assert another.GetFormatted() == "3000 [m]"

    # By default the validation will be performed, and in this cases will raise ValueError.
    another_2 = scalar.CreateCopy(value=FractionValue(5000))
    assert not another_2.IsValid()

    another_3 = units.FractionScalar("another-length", unit="m", value=FractionValue(5000))
    assert not another_3.IsValid()

    # Performing copy between invalid fraction scalars. The validation is not performed on copy.
    copied = another.Copy()
    assert not copied.IsValid()
    assert copied.GetValue("m") == FractionValue(3000)
    assert copied.GetUnit() == "m"
    assert copied.GetFormatted() == "3000 [m]"


def testSetFloatValue() -> None:
    """
    The fraction scalar should accept a float-convertible value. Since the fraction is a subclass
    from scalar, it should be able to respect the same interface.
    """
    f = units.FractionScalar("length", value=FractionValue(0), unit="in")
    f = f.CreateCopy(value=0.75)
    assert f.GetValue("in") == FractionValue(0.75)


def testFractionScalarWithDefaultValueOnCategory(unit_database_len) -> None:
    """
    FractionScalar is not considering the default value from category on initialization
    """
    db = unit_database_len

    db.AddCategory("my length", "length", default_value=FractionValue(5, (1, 2)))

    scalar = units.FractionScalar("my length", unit="m")
    assert scalar.GetValue("m") == FractionValue(5, (1, 2))
    assert scalar.GetUnit() == "m"
    assert scalar.GetFormatted() == "5 1/2 [m]"



from typing import Any
from typing import cast
from typing import Generic
from typing import Iterable
from typing import Iterator
from typing import Optional
from typing import overload
from typing import Sequence
from typing import TypeVar
from typing import Union

from oop_ext.interface import ImplementsInterface

from ._abstractvaluewithquantity import AbstractValueWithQuantityObject
from ._abstractvaluewithquantity import T
from ._quantity import Quantity
from ._scalar import Scalar
from .interfaces import IArray
from .interfaces import ValuesType
from barril._util.types_ import IsNumber
from barril.basic.format_float import FormatFloat
from barril.units.unit_database import CategoryInfo
from barril.units.unit_database import UnitDatabase


__all__ = ["Array"]


SelfT = TypeVar("SelfT", bound="Array")


@ImplementsInterface(IArray)
class Array(AbstractValueWithQuantityObject, Generic[ValuesType]):
    """
    Array represents a sequence of values that also have an unit associated.

    Some ways to construct it (note that usually numpy arrays should be used).

    .. code-block:: python

        Array(numpy.array([0, 1, 2, 3, 4], numpy.float64), "m")

        Array([0, 1, 2, 3, 4], "m")

        Array("length", [0, 1, 2, 3, 4], "m")

        Array(ObtainQuantity("m", "length"), [0, 1, 2, 3, 4])


    Arrays are a ``Generic`` subclass, parametrized by their container type:

    .. code-block:: python

        a = Array([1, 2, 3], "m")
        reveal_type(a)

    .. code-block::

        note: Revealed type is "barril.units._array.Array[builtins.list*[builtins.int*]]"

    Functions/methods which receive arrays can be declared with more specific types:

    .. code-block:: python

        def compute(inputs: Array[np.ndarray]) -> Array[np.ndarray]:
            ...


        def compute(inputs: Array[np.ndarray[np.float64]]) -> Array[np.ndarray[np.float64]]:
            ...

    """

    @overload
    def __init__(self, category: Union[str, Quantity]):
        ...

    @overload
    def __init__(self, values: ValuesType, unit: str, category: Optional[str] = None):
        ...

    @overload
    def __init__(self, category: str, values: ValuesType, unit: str):
        ...

    @overload
    def __init__(self, category: Quantity, values: ValuesType):
        ...

    def __init__(  # type:ignore[misc]
        self, category: str, values: Any = None, unit: Any = None
    ) -> None:
        super().__init__(category, value=values, unit=unit)

    def _InternalCreateWithQuantity(  # type:ignore[override]
        self,
        quantity: Quantity,
        values: Optional[ValuesType] = None,
        unit_database: Optional[UnitDatabase] = None,
        value: Optional[ValuesType] = None,
    ) -> None:
        if value is not None:
            if values is not None:
                raise ValueError("Duplicated values parameter given")
            values = value

        assert values is not None
        self._value = values
        self._unit_database = unit_database or UnitDatabase.GetSingleton()
        self._quantity = quantity
        self._is_valid: Optional[bool] = None
        self._validity_exception: Optional[Exception] = None

    def CheckValidity(self) -> None:
        """
        :raises ValueError: when current value is wrong somehow (out of limits, for example).
        """
        self.ValidateValues(self._value, self._quantity)

    def CreateCopy(  # type:ignore[override]
        self: SelfT,
        values: Optional[ValuesType] = None,
        unit: Optional[str] = None,
        category: Optional[str] = None,
        **kwargs: object
    ) -> SelfT:
        return AbstractValueWithQuantityObject.CreateCopy(
            self, value=values, unit=unit, category=category, **kwargs
        )

    # Values ---------------------------------------------------------------------------------------
    def GetAbstractValue(self, unit: Optional[str] = None) -> ValuesType:
        """
        :param unit: this is the unit in which we want the values
        :returns:
            the values stored. May be an a list of int, float, etc.
        """
        values = self._value
        if unit is None or unit == self._quantity.unit:
            return values

        def IsListOfTuples(v: Any) -> bool:
            try:
                return len(v) > 0 and isinstance(v[0], tuple)
            except TypeError:
                return False  # numpy raises a TypeError if it's a 0D array, so ignores it

        if IsListOfTuples(values):
            result = []
            Convert = self._quantity.Convert
            for elem in values:
                result.append(tuple(Convert(v, unit) for v in elem))
            return type(values)(result)  # type:ignore[call-arg, arg-type]

        else:
            return self._quantity.Convert(values, unit)

    GetValues = GetAbstractValue
    values = property(GetAbstractValue)

    def _GetDefaultValue(
        self, category_info: CategoryInfo, unit: Optional[str] = None
    ) -> ValuesType:
        return cast(ValuesType, [])

    def ValidateValues(self, values: ValuesType, quantity: Quantity) -> None:
        """Set the value to store in this values_quantity. May be an int,
        float, numarray, list of floats, etc.

        :param values:
            The values to be set.

        :param quantity:
            The quantity of the values being passed (note that GetUnit will still return the previous
            unit set -- this unit is only to indicate the internal value).
        """
        if self._is_valid is True:
            return

        if self._validity_exception is not None:
            raise self._validity_exception

        try:
            self._DoValidateValues(values, quantity)
        except Exception as e:
            self._validity_exception = e
            self._is_valid = False
            raise
        else:
            self._is_valid = True

    def _DoValidateValues(self, values: ValuesType, quantity: Quantity) -> None:
        """
        .. seealso:: :meth:`.ValidateValues`
        """
        is_derived = quantity.IsDerived()
        if not is_derived:
            # only check min, max if we have only 1 category (otherwise, we won't have a valid assumption
            # about the actual values)
            category_info = quantity.GetCategoryInfo()

            if category_info.min_value is not None or category_info.max_value is not None:
                # verify if values are in the given limits (if needed)
                CheckValue = quantity.CheckValue
                if len(values) > 0:
                    if isinstance(values[0], tuple):
                        for value in values:
                            if isinstance(value, tuple):
                                for v in value:
                                    CheckValue(v)
                    else:
                        import numpy

                        isnam = numpy.isnan

                        # Search for the first non-NaN value to initialize MIN/MAX.
                        is_numpy = isinstance(values, numpy.ndarray)

                        iterator: Iterator[Any] = iter(values)

                        for value in iterator:
                            if isnam(value):
                                continue

                            # Iterate until we get a non-nan number
                            min_value = max_value = value

                            # Keep on the iteration now that we can already make the check.
                            for value in iterator:
                                if isnam(value):
                                    # NaNs would fail the min_value validation below.
                                    continue

                                if value < min_value:
                                    min_value = value

                                elif value > max_value:
                                    max_value = value

                            if is_numpy:
                                min_value = float(min_value)
                                max_value = float(max_value)

                            CheckValue(min_value)
                            CheckValue(max_value)

                            # Break the outer 'for' used just to get the min/max
                            break

    @classmethod
    def FromScalars(
        cls,
        scalars: Iterable[Scalar],
        *,
        unit: Optional[str] = None,
        category: Optional[str] = None
    ) -> "Array":
        """
        Create an Array from a sequence of Scalars.

        When not defined, the unit and category assumed will be from the first Scalar on the sequence.

        :param values:
            A sequence of Scalars. When the values parameter is an empty sequence and
            the unit is not provided an Array with an empty quantity will be returned.

        :param unit:
            A string representing the unit, if not defined
            the unit from the first Scalar on the sequence will be used.

        :param category:
            A string representing the category, if not defined
            the category from the first Scalar on the sequence will be used.
        """
        scalars = iter(scalars)
        try:
            first_scalar = next(scalars)
        except StopIteration:
            if unit is None and category is None:
                return cls.CreateEmptyArray()
            elif category is None:
                category = UnitDatabase.GetSingleton().GetDefaultCategory(unit)
                return cls(values=[], unit=unit, category=category)  # type:ignore[arg-type]
            else:
                assert unit is None
                return cls(  # type:ignore[call-overload]
                    values=[], unit=unit, category=category
                )  # This actually will raise an exception

        unit = unit or first_scalar.unit
        category = category or first_scalar.category
        values = [first_scalar.GetValue(unit)] + [scalar.GetValue(unit) for scalar in scalars]
        return cls(values=values, unit=unit, category=category)

    @classmethod
    def CreateEmptyArray(cls, values: Optional[Sequence[float]] = None) -> "Array":
        """
        Allows the creation of a array that does not have any associated
        category nor unit.

        :rtype: Array
        """
        if values is None:
            values = []

        quantity = Quantity.CreateEmpty()
        return cls.CreateWithQuantity(quantity, values=values)

    # Equality -------------------------------------------------------------------------------------
    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Array):
            return False

        return (
            tuple(self.values) == tuple(other.values)
            and self._quantity == other._quantity
            and self.unit == other.unit
        )

    def __repr__(self) -> str:
        values_str = "[%s]" % ", ".join(str(v) for v in self.values)
        return "{}({}, {}, {})".format(
            self.__class__.__name__, self.GetQuantityType(), values_str, self.GetUnit()
        )

    def __str__(self) -> str:
        """
        Should return a user-friendly representation of this object.

        :rtype: str
        :returns:
            The formatted string
        """
        if len(self.values) > 0 and isinstance(self.values[0], tuple):
            values_str = " ".join(str(v) for v in self.values)
        else:
            values_str = " ".join((FormatFloat("%g", v)) for v in self.values)

        return values_str + self.GetFormattedSuffix()

    # Basic operators ------------------------------------------------------------------------------
    def __len__(self) -> int:
        return len(self.values)

    @overload
    def __getitem__(self, index: int) -> Any:
        ...

    @overload
    def __getitem__(self, index: slice) -> ValuesType:
        ...

    def __getitem__(self, index: Any) -> Any:
        return self.values[index]

    def __iter__(self) -> Iterator[Any]:
        return iter(self.values)

    def __truediv__(self: SelfT, other: Any) -> SelfT:
        return self._DoOperation(self, other, "Divide")

    def __floordiv__(self: SelfT, other: Any) -> SelfT:
        return self._DoOperation(self, other, "FloorDivide")

    def __mul__(self: SelfT, other: Any) -> SelfT:
        return self._DoOperation(self, other, "Multiply")

    def __add__(self: SelfT, other: Any) -> SelfT:
        return self._DoOperation(self, other, "Sum")

    def __sub__(self: SelfT, other: Any) -> SelfT:
        return self._DoOperation(self, other, "Subtract")

    # Right-Basic operators ------------------------------------------------------------------------
    def __rtruediv__(self: SelfT, other: Any) -> SelfT:
        return self._DoOperation(other, self, "Divide")

    def __rfloordiv__(self: SelfT, other: Any) -> SelfT:
        return self._DoOperation(other, self, "FloorDivide")

    def __rdiv__(self: SelfT, other: Any) -> SelfT:
        return self._DoOperation(other, self, "Divide")

    def __rmul__(self: SelfT, other: Any) -> SelfT:
        return self._DoOperation(other, self, "Multiply")

    def __radd__(self: SelfT, other: Any) -> SelfT:
        return self._DoOperation(other, self, "Sum")

    def __rsub__(self: SelfT, other: Any) -> SelfT:
        return self._DoOperation(other, self, "Subtract")

    def _DoOperation(self, p1: SelfT, p2: SelfT, operation: str) -> SelfT:
        """
        Actually go on and do an operation considering the data we have to transform considering
        any combination of: number, list and numpy
        """
        from ._value_generator import _ValueGenerator
        import numpy

        # get the quantities and setup the value generator properly
        if IsNumber(p1) or isinstance(p1, numpy.ndarray):
            values_iteration = _ValueGenerator(p1, p2.values)
            q2 = p2.GetQuantity()
            q1 = Quantity.CreateEmpty()

        elif IsNumber(p2) or isinstance(p2, numpy.ndarray):
            values_iteration = _ValueGenerator(p1.values, p2)
            q1 = p1.GetQuantity()
            q2 = Quantity.CreateEmpty()

        else:
            values_iteration = _ValueGenerator(p1.values, p2.values)
            q1 = p1.GetQuantity()
            q2 = p2.GetQuantity()

        unit_database = self.GetUnitDatabase()
        operation_func = getattr(unit_database, operation)

        # if handling numpy, just call it all at once!
        if values_iteration.IsNumpy():
            v0, v1 = next(iter(values_iteration))
            q, v = operation_func(q1, q2, v0, v1)
            return self.__class__.CreateWithQuantity(q, v)  # type:ignore[return-value]
        else:
            # not numpy: create a new structure to hold the values
            result = []
            for v0, v1 in values_iteration:
                q, v = operation_func(q1, q2, v0, v1)
                result.append(v)

            if values_iteration.IsTuple():
                result = tuple(result)  # type:ignore[assignment]
            return self.__class__.CreateWithQuantity(q, result)  # type:ignore[return-value]



"""
This module provides the implementation of an Scalar object.
"""
from functools import total_ordering
from typing import Any
from typing import cast
from typing import Optional
from typing import overload
from typing import Tuple
from typing import Union

from oop_ext.interface import ImplementsInterface

from ._abstractvaluewithquantity import AbstractValueWithQuantityObject
from ._quantity import ObtainQuantity
from ._quantity import Quantity
from .interfaces import IQuantity
from .interfaces import IScalar
from .unit_database import CategoryInfo
from .unit_database import UnitDatabase
from barril._util.types_ import IsNumber

__all__ = ["Scalar"]


@total_ordering
@ImplementsInterface(IScalar, IQuantity)
class Scalar(AbstractValueWithQuantityObject):
    """
    This object represents a scalar (a value that has an associated quantity).

    **Scalar Creation**

    Scalars can be created by giving a value, unit or category, in some combinations.

    Assuming the default value for category "length" is 10[m], all the example below create
    equal scalars:

    .. code-block:: python

        Scalar("length")
        Scalar(10.0, "m", "length")
        Scalar(10.0, "m")
        Scalar((10.0, "m"))  # tuple of (value, unit)

    The last form is useful if you want to make a convenient interface for users of a class or method,
    accepting either a tuple or Scalar:

    .. code-block:: python

        def Compute(x, y):
            x, y = Scalar(x), Scalar(y)

    Which allows the user of the method to just pass a tuple, without having to create a Scalar explicitly:

    .. code-block:: python

        Compute(
            x=(10, "m"), y=(15, "m")
        )  # is equivalent to Compute(x=Scalar(10, 'm'), y=Scalar(15, 'm'))

    Note that the following form is invalid, because if category and value is given, unit is mandatory.

    .. code-block:: python

        Scalar("length", 1.0)

    :type _internal_unit: This is the unit in which the value has been set (but not necessarily the
    :ivar _internal_unit:
        unit which is visible to the outside world)

    .. note::
        In case you're wondering, we cannot make a Scalar with __slots__ because our callback
        system (i.e.: callback.After) won't work with it.
    """

    @overload
    def __init__(self, category: Quantity, value: Optional[float] = None) -> None:
        ...

    @overload
    def __init__(self, category: str, value: Optional[float] = None, unit: Optional[str] = None):
        ...

    @overload
    def __init__(self, value: float, unit: str, category: Optional[str] = None):
        ...

    @overload
    def __init__(self, value_and_unit: Tuple[float, str]):
        ...

    def __init__(  # type:ignore[misc]
        self, category: Any, value: Any = None, unit: Any = None
    ) -> None:
        if category.__class__ is tuple:
            # Support for creating a scalar as:
            # Scalar((10, 'm'))
            assert value is None and unit is None
            value, unit = category
            AbstractValueWithQuantityObject.__init__(self, value, unit)
        else:
            AbstractValueWithQuantityObject.__init__(
                self, category=category, value=value, unit=unit
            )

    def _InternalCreateWithQuantity(
        self,
        quantity: Quantity,
        value: Optional[float] = None,
        unit_database: Optional[UnitDatabase] = None,
    ) -> None:
        """
        For internal use only. Is used to initialize the actual quantity.

        :param quantity:
            The quantity of this scalar.

        :param value:
            The initial value.

        :param UnitDatabase unit_database:
            The unit database that should be used for this object.

        """
        if value is None:
            self._value = self._GetDefaultValue(quantity.GetCategoryInfo())
        else:
            self._value = float(value)

        self._unit_database = unit_database or UnitDatabase.GetSingleton()
        self._quantity = quantity

    # Value ----------------------------------------------------------------------------------------
    _value: float

    def GetAbstractValue(self, unit: Optional[str] = None) -> float:
        """

        :param unit:
        """
        if unit is None:
            return self._value
        else:
            return self._quantity.ConvertScalarValue(self._value, unit)

    GetValue = GetAbstractValue
    value = property(GetAbstractValue)

    def _GetDefaultValue(self, category_info: CategoryInfo, unit: Optional[str] = None) -> Any:
        """

        :param category_info:
        :param unit:
        """
        try:
            value = category_info.default_value
        except AttributeError:
            return 0.0

        if unit is not None:
            # needs to convert value to default unit
            value = ObtainQuantity(
                category_info.default_unit, category_info.category
            ).ConvertScalarValue(value, unit)

        return value

    def GetValueAndUnit(self) -> Tuple[float, str]:
        return self._value, self.GetUnit()

    def CheckValidity(self) -> None:
        """
        :raises ValueError: when current value is wrong somehow (out of limits, for example).
        """
        self._quantity.CheckValue(self._value)

    # Handling 'empty' scalar ----------------------------------------------------------------------

    @classmethod
    def CreateEmptyScalar(cls, value: float = 0.0) -> "Scalar":
        """
        Allows the creation of a scalar that does not have any associated
        category nor unit.

        :rtype: Scalar
        :returns:
            Returns an empty scalar (i.e.: without any unit) with the passed value and returns it.
        """
        quantity = Quantity.CreateEmpty()
        return cls.CreateWithQuantity(quantity, value=value)

    # Repr -----------------------------------------------------------------------------------------

    def __repr__(self) -> str:
        return "{}({}, '{}', '{}')".format(
            self.__class__.__name__, self._value, self.GetUnit(), self.GetCategory()
        )

    def __str__(self) -> str:
        """
        Should return a user-friendly representation of this object.

        :rtype: str
        :returns:
            The formatted string
        """
        return self.GetFormatted()

    # Format ---------------------------------------------------------------------------------------

    FORMATTED_VALUE_FORMAT = "%g"

    @classmethod
    def GetFormattedValueFormat(cls) -> str:
        return cls.FORMATTED_VALUE_FORMAT

    @classmethod
    def SetFormattedValueFormat(cls, value_format: str) -> None:
        """
        Sets the format for the numeric part of the scalar.
        """
        try:
            value_format % 1.11
        except TypeError as e:
            raise TypeError(
                "Incompatible format for Scalar value. Expected a format for a float value."
            ) from e
        cls.FORMATTED_VALUE_FORMAT = value_format

    def GetFormattedValue(
        self, unit: Optional[str] = None, value_format: Optional[str] = None
    ) -> str:
        """
        Returns the scalar value formated using the given format or the default format.

        The default format can be modified globally (for all Scalars) using the class method
        SetFormattedFormat.

        .. note:: The unit is NOT returned in this method.

        :param unit:
            The unit in which the value should be gotten.

        :param value_format:
            If not None (default), replaces the default value_format defined in
            Scalar.FORMATTED_VALUE_FORMAT

        :returns:
            A string with the value of this scalar formatted.
        """
        from barril.basic.format_float import FormatFloat

        if value_format is None:
            value_format = self.FORMATTED_VALUE_FORMAT
        return FormatFloat(value_format, self.GetValue(unit))

    def GetFormatted(self, unit: Optional[str] = None, value_format: Optional[str] = None) -> str:
        """
        Returns this scalar in a formatted format (i.e.: formatted value + unit).

        :param unit:
            The unit in which the value should be gotten.


        :param value_format:
            See Scalar.GetFormattedValue
        """
        return self.GetFormattedValue(unit, value_format) + self.GetFormattedSuffix(unit)

    # Compare --------------------------------------------------------------------------------------

    def __eq__(self, other: Any) -> bool:
        return (
            type(self) is type(other)
            and self._value == other.value
            and self._quantity == other._quantity
        )

    def AlmostEqual(self, other: "Scalar", precision: int) -> bool:
        return (
            type(self) is type(other)
            and round(self._value - other.value, precision) == 0
            and self._quantity == other._quantity
        )

    def __hash__(self) -> int:  # type:ignore[override]
        return hash((self._value, self._quantity))

    def __lt__(self, other: Any) -> bool:
        if self.quantity_type != other.quantity_type:
            msg = "can not compare scalars of different quantity types: %r != %r"
            raise TypeError(msg % (self.quantity_type, other.quantity_type))

        v1 = self._value
        v2 = other.GetValue(self.unit)
        return v1 < v2

    # right ----------------------------------------------------------------------------------------
    def __rtruediv__(self, other: Any) -> "Scalar":
        return self._DoOperation(other, self, "Divide", lambda a, b: a / b)

    def __rfloordiv__(self, other: Any) -> "Scalar":
        return self._DoOperation(other, self, "FloorDivide", lambda a, b: a // b)

    def __rmul__(self, other: Any) -> "Scalar":
        return self._DoOperation(other, self, "Multiply", lambda a, b: a * b)

    def __rsub__(self, other: Any) -> "Scalar":
        return self._DoOperation(other, self, "Subtract", lambda a, b: a - b)

    def __radd__(self, other: Any) -> "Scalar":
        return self._DoOperation(other, self, "Sum", lambda a, b: a + b)

    # basic ----------------------------------------------------------------------------------------
    def __truediv__(self, other: Any) -> "Scalar":
        return self._DoOperation(self, other, "Divide", lambda a, b: a / b)

    def __floordiv__(self, other: Any) -> "Scalar":
        return self._DoOperation(self, other, "FloorDivide", lambda a, b: a // b)

    def __mul__(self, other: Any) -> "Scalar":
        return self._DoOperation(self, other, "Multiply", lambda a, b: a * b)

    def __sub__(self, other: Any) -> "Scalar":
        return self._DoOperation(self, other, "Subtract", lambda a, b: a - b)

    def __add__(self, other: Any) -> "Scalar":
        return self._DoOperation(self, other, "Sum", lambda a, b: a + b)

    def __pow__(self, exponent: int) -> "Scalar":
        result = self
        for _ in range(exponent - 1):
            result = result * self
        return result

    def _DoOperation(self, p1: Any, p2: Any, operation: Any, callback_operation: Any) -> Any:
        p1_is_number = IsNumber(p1)
        if p1_is_number and operation not in ["Divide", "FloorDivide"]:
            return self.__class__.CreateWithQuantity(
                self._quantity, callback_operation(p1, self._value)
            )

        if IsNumber(p2):
            return self.__class__.CreateWithQuantity(
                self._quantity, callback_operation(self._value, p2)
            )

        unit_database = self._unit_database
        operation_func = getattr(unit_database, operation)
        if p1_is_number:
            assert operation in [
                "Divide",
                "FloorDivide",
            ], "Only operation Divide and FloorDivide allowed here!"
            q, v = operation_func(Quantity.CreateEmpty(), p2.GetQuantity(), p1, p2.value)
        else:
            q, v = operation_func(p1.GetQuantity(), p2.GetQuantity(), self._value, p2.value)
        return self.__class__.CreateWithQuantity(q, v)

    def __reduce__(self) -> Any:
        """
        Defining reduce so that we can pickle scalars.
        """
        return Scalar, (self._quantity, self.value, None)  # Unit defined in quantity



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from barril.units import Quantity

# constants for unknown quantities
UNKNOWN_QUANTITY_TYPE = "Unknown"
UNKNOWN_UNIT = "<unknown>"

LENGTH_QUANTITY_TYPE = "length"


def CreateUnknwonwReadOnlyQuantity() -> "Quantity":
    """
    :returns:
        Returns a read only quantity for an unknown quantity.
    """
    from ._quantity import ObtainQuantity

    return ObtainQuantity(UNKNOWN_UNIT, UNKNOWN_QUANTITY_TYPE)
